# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-08-15 15:17
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('active', models.BooleanField(default=True, help_text='Users may only charge to an account if it is active. Deactivate the account to block future billable activity (such as tool usage and consumable check-outs) of all the projects that belong to it.')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ActivityHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('action', models.BooleanField(choices=[(True, 'Activated'), (False, 'Deactivated')], default=None, help_text='The target state (activated or deactivated).')),
                ('date', models.DateTimeField(default=django.utils.timezone.now, help_text='The time at which the active state was changed.')),
            ],
            options={
                'verbose_name_plural': 'activity histories',
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='Alert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(blank=True, max_length=100)),
                ('contents', models.CharField(max_length=500)),
                ('creation_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('debut_time', models.DateTimeField(help_text='The alert will not be displayed to users until the debut time is reached.')),
                ('expiration_time', models.DateTimeField(blank=True, help_text='The alert can be deleted after the expiration time is reached.', null=True)),
                ('dismissible', models.BooleanField(default=False, help_text="Allows the user to delete the alert. This is only valid when the 'user' field is set.")),
            ],
            options={
                'ordering': ['-debut_time'],
            },
        ),
        migrations.CreateModel(
            name='Area',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='What is the name of this area? The name will be displayed on the tablet login and logout pages.', max_length=200)),
                ('welcome_message', models.TextField(help_text='The welcome message will be displayed on the tablet login page. You can use HTML and JavaScript.')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='AreaAccessRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(default=django.utils.timezone.now)),
                ('end', models.DateTimeField(blank=True, null=True)),
                ('area', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area')),
            ],
            options={
                'ordering': ['-start'],
            },
        ),
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('expiration_date', models.DateTimeField(blank=True, help_text='The comment will only be visible until this date.', null=True)),
                ('visible', models.BooleanField(default=True)),
                ('hide_date', models.DateTimeField(blank=True, help_text='The date when this comment was hidden. If it is still visible or has expired then this date should be empty.', null=True)),
                ('content', models.TextField()),
            ],
            options={
                'ordering': ['-creation_date'],
            },
        ),
        migrations.CreateModel(
            name='Configuration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of this overall configuration. This text is displayed as a label on the tool control page.', max_length=200)),
                ('configurable_item_name', models.CharField(blank=True, help_text='The name of the tool part being configured. This text is displayed as a label on the tool control page. Leave this field blank if there is only one configuration slot.', max_length=200, null=True)),
                ('advance_notice_limit', models.PositiveIntegerField(help_text='Configuration changes must be made this many hours in advance.')),
                ('display_priority', models.PositiveIntegerField(help_text='The order in which this configuration will be displayed beside others when making a reservation and controlling a tool. Can be any positive integer including 0. Lower values are displayed first.')),
                ('prompt', models.TextField(blank=True, help_text='The textual description the user will see when making a configuration choice.', null=True)),
                ('current_settings', models.TextField(blank=True, help_text='The current configuration settings for a tool. Multiple values are separated by commas.', null=True)),
                ('available_settings', models.TextField(blank=True, help_text='The available choices to select for this configuration option. Multiple values are separated by commas.', null=True)),
                ('qualified_users_are_maintainers', models.BooleanField(default=False, help_text='Any user that is qualified to use the tool that this configuration applies to may also change this configuration. Checking this box implicitly adds qualified users to the maintainers list.')),
                ('exclude_from_configuration_agenda', models.BooleanField(default=False, help_text="Reservations containing this configuration will be excluded from the NanoFab technician's Configuration Agenda page.")),
                ('absence_string', models.CharField(blank=True, help_text='The text that appears to indicate absence of a choice.', max_length=100, null=True)),
            ],
            options={
                'ordering': ['tool', 'name'],
            },
        ),
        migrations.CreateModel(
            name='ConfigurationHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('modification_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('slot', models.PositiveIntegerField()),
                ('setting', models.TextField()),
                ('configuration', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Configuration')),
            ],
            options={
                'verbose_name_plural': 'Configuration histories',
                'ordering': ['-modification_time'],
            },
        ),
        migrations.CreateModel(
            name='Consumable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('quantity', models.IntegerField(help_text='The number of items currently in stock.')),
                ('reminder_threshold', models.IntegerField(help_text='More of this item should be ordered when the quantity falls below this threshold.')),
                ('reminder_email', models.EmailField(help_text='An email will be sent to this address when the quantity of this item falls below the reminder threshold.', max_length=254)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ConsumableCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'verbose_name_plural': 'Consumable categories',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ConsumableWithdraw',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField()),
                ('date', models.DateTimeField(default=django.utils.timezone.now, help_text='The date and time when the user withdrew the consumable.')),
                ('consumable', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Consumable')),
            ],
            options={
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='ContactInformation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('image', models.ImageField(blank=True, help_text='Portraits are resized to 266 pixels high and 200 pixels wide. Crop portraits to these dimensions before uploading for optimal bandwidth usage', upload_to='')),
                ('email', models.EmailField(blank=True, max_length=254)),
                ('office_phone', models.CharField(blank=True, max_length=40)),
                ('office_location', models.CharField(blank=True, max_length=200)),
                ('mobile_phone', models.CharField(blank=True, max_length=40)),
                ('mobile_phone_is_sms_capable', models.BooleanField(default=True, help_text="Is the mobile phone capable of receiving text messages? If so, a link will be displayed for users to click to send a text message to the recipient when viewing the 'Contact information' page.", verbose_name='Mobile phone is SMS capable')),
            ],
            options={
                'verbose_name_plural': 'Contact information',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ContactInformationCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('display_order', models.IntegerField(help_text="Contact information categories are sorted according to display order. The lowest value category is displayed first in the 'Contact information' page.")),
            ],
            options={
                'verbose_name_plural': 'Contact information categories',
                'ordering': ['display_order', 'name'],
            },
        ),
        migrations.CreateModel(
            name='Customization',
            fields=[
                ('name', models.CharField(max_length=50, primary_key=True, serialize=False)),
                ('value', models.TextField()),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Door',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('area', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='doors', to='NEMO.Area')),
            ],
        ),
        migrations.CreateModel(
            name='Interlock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('channel', models.PositiveIntegerField()),
                ('state', models.IntegerField(choices=[(-1, 'Unknown'), (1, 'Unlocked'), (2, 'Locked')], default=-1)),
                ('most_recent_reply', models.TextField(default='None')),
            ],
            options={
                'ordering': ['card__server', 'card__number', 'channel'],
            },
        ),
        migrations.CreateModel(
            name='InterlockCard',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('server', models.CharField(max_length=100)),
                ('port', models.PositiveIntegerField()),
                ('number', models.PositiveIntegerField()),
                ('even_port', models.PositiveIntegerField()),
                ('odd_port', models.PositiveIntegerField()),
            ],
            options={
                'ordering': ['server', 'number'],
            },
        ),
        migrations.CreateModel(
            name='LandingPageChoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(help_text='An image that symbolizes the choice. It is automatically resized to 128x128 pixels when displayed, so set the image to this size before uploading to optimize bandwidth usage and landing page load time', upload_to='')),
                ('name', models.CharField(help_text='The textual name that will be displayed underneath the image', max_length=40)),
                ('url', models.CharField(help_text='The URL that the choice leads to when clicked. Relative paths such as /calendar/ are used when linking within the NEMO site. Use fully qualified URL paths such as https://www.google.com/ to link to external sites.', max_length=200, verbose_name='URL')),
                ('display_priority', models.IntegerField(help_text='The order in which choices are displayed on the landing page, from left to right, top to bottom. Lower values are displayed first.')),
                ('open_in_new_tab', models.BooleanField(default=False, help_text="Open the URL in a new browser tab when it's clicked")),
                ('secure_referral', models.BooleanField(default=True, help_text="Improves security by blocking HTTP referer [sic] information from the targeted page. Enabling this prevents the target page from manipulating the calling page's DOM with JavaScript. This should always be used for external links. It is safe to uncheck this when linking within the NEMO site. Leave this box checked if you don't know what this means")),
                ('hide_from_mobile_devices', models.BooleanField(default=False, help_text='Hides this choice when the landing page is viewed from a mobile device')),
                ('hide_from_desktop_computers', models.BooleanField(default=False, help_text='Hides this choice when the landing page is viewed from a desktop computer')),
            ],
            options={
                'ordering': ['display_priority'],
            },
        ),
        migrations.CreateModel(
            name='MembershipHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('parent_object_id', models.PositiveIntegerField()),
                ('child_object_id', models.PositiveIntegerField()),
                ('date', models.DateTimeField(default=django.utils.timezone.now, help_text='The time at which the membership status was changed.')),
                ('action', models.BooleanField(choices=[(True, 'Added'), (False, 'Removed')], default=None)),
            ],
            options={
                'verbose_name_plural': 'membership histories',
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='PhysicalAccessLevel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('schedule', models.IntegerField(choices=[(0, 'Always'), (1, 'Weekdays, 7am to midnight'), (2, 'Weekends')])),
                ('area', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='PhysicalAccessLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time', models.DateTimeField()),
                ('result', models.BooleanField(choices=[(False, 'Deny'), (True, 'Allow')], default=None)),
                ('details', models.TextField(blank=True, help_text='Any details that should accompany the log entry. For example, the reason physical access was denied.', null=True)),
                ('door', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Door')),
            ],
            options={
                'ordering': ['-time'],
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('application_identifier', models.CharField(max_length=100)),
                ('active', models.BooleanField(default=True, help_text='Users may only charge to a project if it is active. Deactivate the project to block billable activity (such as tool usage and consumable check-outs).')),
                ('account', models.ForeignKey(help_text='All charges for this project will be billed to the selected account.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Account')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Reservation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('start', models.DateTimeField(verbose_name='start')),
                ('end', models.DateTimeField(verbose_name='end')),
                ('short_notice', models.BooleanField(default=None, help_text='Indicates that the reservation was made after the configuration deadline for a tool. NanoFab staff may not have enough time to properly configure the tool before the user is scheduled to use it.')),
                ('cancelled', models.BooleanField(default=False, help_text='Indicates that the reservation has been cancelled, moved, or resized.')),
                ('cancellation_time', models.DateTimeField(blank=True, null=True)),
                ('missed', models.BooleanField(default=False, help_text='Indicates that the tool was not enabled by anyone before the tool\'s "missed reservation threshold" passed.')),
                ('shortened', models.BooleanField(default=False, help_text='Indicates that the user finished using the tool and relinquished the remaining time on their reservation. The reservation will no longer be visible on the calendar and a descendant reservation will be created in place of the existing one.')),
                ('additional_information', models.TextField(blank=True, null=True)),
                ('self_configuration', models.BooleanField(default=False, help_text='When checked, indicates that the user will perform their own tool configuration (instead of requesting that the NanoFab staff configure it for them).')),
                ('title', models.TextField(blank=True, default='', help_text="Shows a custom title for this reservation on the calendar. Leave this field blank to display the reservation's user name as the title (which is the default behaviour).", max_length=200)),
            ],
            options={
                'ordering': ['-start'],
            },
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('available', models.BooleanField(default=True, help_text='Indicates whether the resource is available to be used.')),
                ('restriction_message', models.TextField(help_text='The message that is displayed to users on the tool control page when this resource is unavailable.')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ResourceCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
            ],
            options={
                'verbose_name_plural': 'resource categories',
            },
        ),
        migrations.CreateModel(
            name='SafetyIssue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=200)),
                ('creation_time', models.DateTimeField(auto_now_add=True)),
                ('visible', models.BooleanField(default=True, help_text='Should this safety issue be visible to all users? When unchecked, the issue is only visible to staff.')),
                ('concern', models.TextField()),
                ('progress', models.TextField(blank=True, null=True)),
                ('resolution', models.TextField(blank=True, null=True)),
                ('resolved', models.BooleanField(default=False)),
                ('resolution_time', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'ordering': ['-creation_time'],
            },
        ),
        migrations.CreateModel(
            name='StaffCharge',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(default=django.utils.timezone.now)),
                ('end', models.DateTimeField(blank=True, null=True)),
                ('validated', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ['-start'],
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(-1, 'Cancelled'), (0, 'Requires attention'), (1, 'Work in progress'), (2, 'Complete')], default=0)),
                ('urgency', models.IntegerField(choices=[(-1, 'Low'), (0, 'Normal'), (1, 'High')])),
                ('force_shutdown', models.BooleanField(default=None, help_text='Indicates that the tool this task relates to will be shutdown until the task is resolved.')),
                ('safety_hazard', models.BooleanField(default=None, help_text='Indicates that this task represents a safety hazard to the NanoFab.')),
                ('creation_time', models.DateTimeField(default=django.utils.timezone.now, help_text='The date and time when the task was created.')),
                ('problem_description', models.TextField(blank=True, null=True)),
                ('first_response_time', models.DateTimeField(blank=True, help_text='The timestamp of when a staff member initially responds to the task by changing its status.', null=True)),
                ('progress_description', models.TextField(blank=True, null=True)),
                ('last_updated', models.DateTimeField(blank=True, help_text='The last time this task was modified. (Creating the task does not count as modifying it.)', null=True)),
                ('estimated_resolution_time', models.DateTimeField(blank=True, help_text='The estimated date and time that the task will be resolved.', null=True)),
                ('resolution_time', models.DateTimeField(blank=True, help_text='The timestamp of when the task was marked complete or cancelled.', null=True)),
                ('resolution_description', models.TextField(blank=True, null=True)),
            ],
            options={
                'ordering': ['-creation_time'],
            },
        ),
        migrations.CreateModel(
            name='TaskCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('stage', models.IntegerField(choices=[(0, 'Initial assessment'), (1, 'Completion')])),
            ],
            options={
                'verbose_name_plural': 'Task categories',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Tool',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('category', models.CharField(help_text='Create sub-categories using slashes. For example "Category 1/Sub-category 1".', max_length=1000)),
                ('visible', models.BooleanField(default=True, help_text='Specifies whether this tool is visible to users.')),
                ('operational', models.BooleanField(default=False, help_text='Marking the tool non-operational will prevent users from using the tool.')),
                ('location', models.CharField(max_length=100)),
                ('phone_number', models.CharField(max_length=100)),
                ('notification_email_address', models.EmailField(blank=True, help_text='Messages that relate to this tool (such as comments, problems, and shutdowns) will be forwarded to this email address. This can be a normal email address or a mailing list address.', max_length=254, null=True)),
                ('reservation_horizon', models.PositiveIntegerField(blank=True, default=14, help_text='Users may create reservations this many days in advance. Leave this field blank to indicate that no reservation horizon exists for this tool.', null=True)),
                ('minimum_usage_block_time', models.PositiveIntegerField(blank=True, help_text='The minimum amount of time (in minutes) that a user must reserve this tool for a single reservation. Leave this field blank to indicate that no minimum usage block time exists for this tool.', null=True)),
                ('maximum_usage_block_time', models.PositiveIntegerField(blank=True, help_text='The maximum amount of time (in minutes) that a user may reserve this tool for a single reservation. Leave this field blank to indicate that no maximum usage block time exists for this tool.', null=True)),
                ('maximum_reservations_per_day', models.PositiveIntegerField(blank=True, help_text='The maximum number of reservations a user may make per day for this tool.', null=True)),
                ('minimum_time_between_reservations', models.PositiveIntegerField(blank=True, help_text='The minimum amount of time (in minutes) that the same user must have between any two reservations for this tool.', null=True)),
                ('maximum_future_reservation_time', models.PositiveIntegerField(blank=True, help_text='The maximum amount of time (in minutes) that a user may reserve from the current time onwards.', null=True)),
                ('missed_reservation_threshold', models.PositiveIntegerField(blank=True, help_text='The amount of time (in minutes) that a tool reservation may go unused before it is automatically marked as "missed" and hidden from the calendar. Usage can be from any user, regardless of who the reservation was originally created for. The cancellation process is triggered by a timed job on the web server.', null=True)),
                ('allow_delayed_logoff', models.BooleanField(default=False, help_text='Upon logging off users may enter a delay before another user may use the tool. Some tools require "spin-down" or cleaning time after use.')),
                ('interlock', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='NEMO.Interlock')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='TrainingSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('duration', models.PositiveIntegerField(help_text='The duration of the training session in minutes.')),
                ('type', models.IntegerField(choices=[(0, 'Individual'), (1, 'Group')])),
                ('date', models.DateTimeField(default=django.utils.timezone.now)),
                ('qualified', models.BooleanField(default=False, help_text='Indicates that after this training session the user was qualified to use the tool.')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project')),
                ('tool', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool')),
            ],
            options={
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='UsageEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(default=django.utils.timezone.now)),
                ('end', models.DateTimeField(blank=True, null=True)),
                ('validated', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ['-start'],
            },
        ),
        migrations.CreateModel(
            name='UserType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, unique=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(max_length=100, unique=True)),
                ('first_name', models.CharField(max_length=100)),
                ('last_name', models.CharField(max_length=100)),
                ('email', models.EmailField(max_length=254, verbose_name='email address')),
                ('domain', models.CharField(blank=True, help_text='The Active Directory domain that the account resides on', max_length=100)),
                ('badge_number', models.PositiveIntegerField(blank=True, help_text='The badge number associated with this user. This number must correctly correspond to a user in order for the tablet-login system (in the NanoFab lobby) to work properly.', null=True, unique=True)),
                ('access_expiration', models.DateField(blank=True, help_text='The user will lose all access rights after this date. Typically this is used to ensure that safety training has been completed by the user every year.', null=True)),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user can log in to NEMO. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_technician', models.BooleanField(default=False, help_text='Specifies how to bill staff time for this user. When checked, customers are billed at technician rates.', verbose_name='technician status')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('training_required', models.BooleanField(default=True, help_text='When selected, the user is blocked from all reservation and tool usage capabilities.')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now)),
                ('last_login', models.DateTimeField(blank=True, null=True)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of his/her group.', to='auth.Group')),
                ('physical_access_levels', models.ManyToManyField(blank=True, related_name='users', to='NEMO.PhysicalAccessLevel')),
                ('projects', models.ManyToManyField(blank=True, help_text='Select the projects that this user is currently working on.', to='NEMO.Project')),
                ('qualifications', models.ManyToManyField(blank=True, help_text='Select the tools that the user is qualified to use.', to='NEMO.Tool')),
                ('type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='NEMO.UserType')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', to='auth.Permission')),
            ],
            options={
                'ordering': ['first_name'],
                'permissions': (('trigger_timed_services', 'Can trigger timed services'), ('use_billing_api', 'Can use billing API'), ('kiosk', 'Kiosk services')),
            },
        ),
        migrations.AddField(
            model_name='usageevent',
            name='operator',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='usage_event_operator', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='usageevent',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project'),
        ),
        migrations.AddField(
            model_name='usageevent',
            name='tool',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='usageevent',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='usage_event_user', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='trainingsession',
            name='trainee',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='student_set', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='trainingsession',
            name='trainer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='teacher_set', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='tool',
            name='primary_owner',
            field=models.ForeignKey(help_text='The staff member who is responsible for administration of this tool.', on_delete=django.db.models.deletion.CASCADE, related_name='primary_tool_owner', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='tool',
            name='requires_area_access',
            field=models.ForeignKey(blank=True, help_text='Indicates that this tool is physically located in a billable area and requires an active area access record in order to be operated.', null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area'),
        ),
        migrations.AddField(
            model_name='tool',
            name='secondary_owner',
            field=models.ForeignKey(help_text='The alternate staff member who is responsible for administration of this tool.', on_delete=django.db.models.deletion.CASCADE, related_name='secondary_tool_owner', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='creator',
            field=models.ForeignKey(help_text='The user who created the task.', on_delete=django.db.models.deletion.CASCADE, related_name='created_tasks', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='first_responder',
            field=models.ForeignKey(blank=True, help_text='The staff member who initially assessed the task after it was reported.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='task_first_responder', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='last_updated_by',
            field=models.ForeignKey(blank=True, help_text='The last user who modified this task. This should always be a staff member.', null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='problem_category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='problem_category', to='NEMO.TaskCategory'),
        ),
        migrations.AddField(
            model_name='task',
            name='resolution_category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='resolution_category', to='NEMO.TaskCategory'),
        ),
        migrations.AddField(
            model_name='task',
            name='resolver',
            field=models.ForeignKey(blank=True, help_text='The staff member who resolved the task.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='task_resolver', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='tool',
            field=models.ForeignKey(help_text='The tool that this task relates to.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='staffcharge',
            name='customer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='staff_charge_customer', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='staffcharge',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project'),
        ),
        migrations.AddField(
            model_name='staffcharge',
            name='staff_member',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='staff_charge_actor', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='safetyissue',
            name='reporter',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='reported_safety_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='safetyissue',
            name='resolver',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='resolved_safety_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='resource',
            name='category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.ResourceCategory'),
        ),
        migrations.AddField(
            model_name='resource',
            name='dependent_areas',
            field=models.ManyToManyField(blank=True, help_text='Users will not be able to login to these areas when the resource is unavailable.', related_name='required_resources', to='NEMO.Area'),
        ),
        migrations.AddField(
            model_name='resource',
            name='fully_dependent_tools',
            field=models.ManyToManyField(blank=True, help_text='These tools will be completely inoperable if the resource is unavailable.', related_name='required_resource_set', to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='resource',
            name='partially_dependent_tools',
            field=models.ManyToManyField(blank=True, help_text='These tools depend on this resource but can operated at a reduced capacity if the resource is unavailable.', related_name='nonrequired_resource_set', to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='reservation',
            name='cancelled_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='reservation',
            name='creator',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reservation_creator', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='reservation',
            name='descendant',
            field=models.OneToOneField(blank=True, help_text='Any time a reservation is moved or resized, the old reservation is cancelled and a new reservation with updated information takes its place. This field links the old reservation to the new one, so the history of reservation moves & changes can be easily tracked.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='ancestor', to='NEMO.Reservation'),
        ),
        migrations.AddField(
            model_name='reservation',
            name='project',
            field=models.ForeignKey(blank=True, help_text='Indicates the intended project for this reservation. A missed reservation would be billed to this project.', null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project'),
        ),
        migrations.AddField(
            model_name='reservation',
            name='tool',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='reservation',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reservation_user', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='physicalaccesslog',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='membershiphistory',
            name='authorizer',
            field=models.ForeignKey(help_text='The staff member who changed the membership status of the account, project, or user in question.', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='membershiphistory',
            name='child_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='child_content_type', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='membershiphistory',
            name='parent_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='parent_content_type', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='interlock',
            name='card',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.InterlockCard'),
        ),
        migrations.AddField(
            model_name='door',
            name='interlock',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Interlock'),
        ),
        migrations.AddField(
            model_name='contactinformation',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.ContactInformationCategory'),
        ),
        migrations.AddField(
            model_name='consumablewithdraw',
            name='customer',
            field=models.ForeignKey(help_text='The user who will use the consumable item.', on_delete=django.db.models.deletion.CASCADE, related_name='consumable_user', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='consumablewithdraw',
            name='merchant',
            field=models.ForeignKey(help_text='The staff member that performed the withdraw.', on_delete=django.db.models.deletion.CASCADE, related_name='consumable_merchant', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='consumablewithdraw',
            name='project',
            field=models.ForeignKey(help_text='The withdraw will be billed to this project.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project'),
        ),
        migrations.AddField(
            model_name='consumable',
            name='category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.ConsumableCategory'),
        ),
        migrations.AddField(
            model_name='configurationhistory',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='configuration',
            name='maintainers',
            field=models.ManyToManyField(blank=True, help_text='Select the users that are allowed to change this configuration.', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='configuration',
            name='tool',
            field=models.ForeignKey(help_text='The tool that this configuration option applies to.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='comment',
            name='author',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='comment',
            name='hidden_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='hidden_comments', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='comment',
            name='tool',
            field=models.ForeignKey(help_text='The tool that this comment relates to.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='areaaccessrecord',
            name='customer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='areaaccessrecord',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Project'),
        ),
        migrations.AddField(
            model_name='areaaccessrecord',
            name='staff_charge',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.StaffCharge'),
        ),
        migrations.AddField(
            model_name='alert',
            name='creator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='alert',
            name='user',
            field=models.ForeignKey(blank=True, help_text='The alert will be visible for this user. The alert is visible to all users when this is empty.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='alerts', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='activityhistory',
            name='authorizer',
            field=models.ForeignKey(help_text='The staff member who changed the active state of the account, project, or user in question.', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='activityhistory',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterUniqueTogether(
            name='interlock',
            unique_together=set([('card', 'channel')]),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.decorators import disable_session_expiry_refresh
from NEMO.views.status_dashboard import create_tool_summary


@login_required
@require_GET
@disable_session_expiry_refresh
def refresh_sidebar_icons(request):
	tool_summary = create_tool_summary()
	return render(request, 'refresh_sidebar_icons.html', {'tool_summary': tool_summary})



from django_filters import FilterSet, IsoDateTimeFilter, BooleanFilter
from django_filters.widgets import BooleanWidget

from NEMO.models import Reservation, UsageEvent, AreaAccessRecord, User


class ReservationFilter(FilterSet):
	start_gte = IsoDateTimeFilter('start', lookup_expr='gte')
	start_lt = IsoDateTimeFilter('start', lookup_expr='lt')
	missed = BooleanFilter('missed', widget=BooleanWidget())

	class Meta:
		model = Reservation
		fields = []


class UsageEventFilter(FilterSet):
	start_gte = IsoDateTimeFilter('start', lookup_expr='gte')
	start_lt = IsoDateTimeFilter('start', lookup_expr='lt')

	class Meta:
		model = UsageEvent
		fields = []


class AreaAccessRecordFilter(FilterSet):
	start_gte = IsoDateTimeFilter('start', lookup_expr='gte')
	start_lt = IsoDateTimeFilter('start', lookup_expr='lt')

	class Meta:
		model = AreaAccessRecord
		fields = []


class UserFilter(FilterSet):

	class Meta:
		model = User
		fields = {
			'date_joined': ['month', 'year'],
		}



from copy import deepcopy
from datetime import timedelta
from http import HTTPStatus

from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import Project, Reservation, Tool, UsageEvent, User
from NEMO.views.policy import check_policy_to_disable_tool, check_policy_to_enable_tool
from NEMO.views.status_dashboard import create_tool_summary
from NEMO.utilities import quiet_int
from NEMO.widgets.dynamic_form import DynamicForm


@login_required
@permission_required('NEMO.kiosk')
@require_POST
def enable_tool(request):
	tool = Tool.objects.get(id=request.POST['tool_id'])
	customer = User.objects.get(id=request.POST['customer_id'])
	project = Project.objects.get(id=request.POST['project_id'])

	response = check_policy_to_enable_tool(tool, operator=customer, user=customer, project=project, staff_charge=False)
	if response.status_code != HTTPStatus.OK:
		dictionary = {
			'message': 'You are not authorized to enable this tool. {}'.format(response),
			'delay': 10,
		}
		return render(request, 'kiosk/acknowledgement.html', dictionary)

	# All policy checks passed so enable the tool for the user.
	if tool.interlock and not tool.interlock.unlock():
		raise Exception("The interlock command for this tool failed. The error message returned: " + str(tool.interlock.most_recent_reply))

	# Create a new usage event to track how long the user uses the tool.
	new_usage_event = UsageEvent()
	new_usage_event.operator = customer
	new_usage_event.user = customer
	new_usage_event.project = project
	new_usage_event.tool = tool
	new_usage_event.save()

	dictionary = {'message': 'You can now use the {}'.format(tool)}
	return render(request, 'kiosk/acknowledgement.html', dictionary)


@login_required
@permission_required('NEMO.kiosk')
@require_POST
def disable_tool(request):
	tool = Tool.objects.get(id=request.POST['tool_id'])
	customer = User.objects.get(id=request.POST['customer_id'])
	downtime = timedelta(minutes=quiet_int(request.POST.get('downtime')))
	response = check_policy_to_disable_tool(tool, customer, downtime)
	if response.status_code != HTTPStatus.OK:
		dictionary = {
			'message': response.content,
			'delay': 10,
		}
		return render(request, 'kiosk/acknowledgement.html', dictionary)
	try:
		current_reservation = Reservation.objects.get(start__lt=timezone.now(), end__gt=timezone.now(), cancelled=False, missed=False, shortened=False, user=customer, tool=tool)
		# Staff are exempt from mandatory reservation shortening when tool usage is complete.
		if customer.is_staff is False:
			# Shorten the user's reservation to the current time because they're done using the tool.
			new_reservation = deepcopy(current_reservation)
			new_reservation.id = None
			new_reservation.pk = None
			new_reservation.end = timezone.now()
			new_reservation.save()
			current_reservation.shortened = True
			current_reservation.descendant = new_reservation
			current_reservation.save()
	except Reservation.DoesNotExist:
		pass

	# All policy checks passed so disable the tool for the user.
	if tool.interlock and not tool.interlock.lock():
		raise Exception("The interlock command for this tool failed. The error message returned: " + str(tool.interlock.most_recent_reply))
	# End the current usage event for the tool and save it.
	current_usage_event = tool.get_current_usage_event()
	current_usage_event.end = timezone.now() + downtime

	# Collect post-usage questions
	current_usage_event.run_data = DynamicForm(tool.post_usage_questions).extract(request)
	current_usage_event.save()

	dictionary = {'message': 'You are no longer using the {}'.format(tool)}
	return render(request, 'kiosk/acknowledgement.html', dictionary)


@login_required
@permission_required('NEMO.kiosk')
@require_GET
def choices(request):
	try:
		customer = User.objects.get(badge_number=request.GET['badge_number'])
	except:
		dictionary = {'message': "Your badge wasn't recognized. If you got a new one recently then we'll need to update your account. Please visit the NanoFab user office to resolve the problem."}
		return render(request, 'kiosk/acknowledgement.html', dictionary)
	dictionary = {
		'customer': customer,
		'usage_events': UsageEvent.objects.filter(operator=customer.id, end=None).order_by('tool__name').prefetch_related('tool', 'project'),
		'tools': Tool.objects.filter(visible=True, location=request.GET['location']),
		'tool_summary': create_tool_summary(),
		'categories': [t[0] for t in Tool.objects.filter(visible=True).order_by('category').values_list('category').distinct()],
	}
	return render(request, 'kiosk/choices.html', dictionary)


@login_required
@permission_required('NEMO.kiosk')
@require_GET
def category_choices(request, category, user_id):
	try:
		customer = User.objects.get(id=user_id)
	except:
		dictionary = {'message': "Your badge wasn't recognized. If you got a new one recently then we'll need to update your account. Please visit the NanoFab user office to resolve the problem."}
		return render(request, 'kiosk/acknowledgement.html', dictionary)
	tools = Tool.objects.filter(visible=True, category=category)
	dictionary = {
		'customer': customer,
		'category': category,
		'tools': tools,
		'tool_summary': create_tool_summary(),
	}
	return render(request, 'kiosk/category_choices.html', dictionary)


@login_required
@permission_required('NEMO.kiosk')
@require_GET
def tool_information(request, tool_id, user_id, back):
	tool = Tool.objects.get(id=tool_id, visible=True)
	customer = User.objects.get(id=user_id)
	dictionary = {
		'customer': customer,
		'tool': tool,
		'rendered_configuration_html': tool.configuration_widget(customer),
		'post_usage_questions': DynamicForm(tool.post_usage_questions).render(),
		'back': back,
	}
	try:
		current_reservation = Reservation.objects.get(start__lt=timezone.now(), end__gt=timezone.now(), cancelled=False, missed=False, shortened=False, user=customer, tool=tool)
		remaining_reservation_duration = int((current_reservation.end - timezone.now()).total_seconds() / 60)
		# We don't need to bother telling the user their reservation will be shortened if there's less than two minutes left.
		# Staff are exempt from reservation shortening.
		if remaining_reservation_duration > 2 and not customer.is_staff:
			dictionary['remaining_reservation_duration'] = remaining_reservation_duration
	except Reservation.DoesNotExist:
		pass
	return render(request, 'kiosk/tool_information.html', dictionary)


@login_required
@permission_required('NEMO.kiosk')
@require_GET
def kiosk(request, location=None):
	if location and Tool.objects.filter(location=location, visible=True).exists():
		dictionary = {
			'location': location,
		}
		return render(request, 'kiosk/kiosk.html', dictionary)
	else:
		locations = sorted(list(set([tool.location for tool in Tool.objects.filter(visible=True)])))
		dictionary = {
			'locations': [{'url': reverse('kiosk', kwargs={'location': location}), 'name': location} for location in locations]
		}
		return render(request, 'kiosk/location_directory.html', dictionary)



from django.contrib.admin.views.decorators import staff_member_required
from django.db.models import F, Q
from django.http import HttpResponse
from django.shortcuts import render, get_object_or_404
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import UsageEvent, StaffCharge, User
from NEMO.utilities import month_list, get_month_timeframe


@staff_member_required(login_url=None)
@require_GET
def remote_work(request):
	first_of_the_month, last_of_the_month = get_month_timeframe(request.GET.get('date'))
	operator = request.GET.get('operator')
	if operator:
		if operator == "all staff":
			operator = None
		else:
			operator = get_object_or_404(User, id=operator)
	else:
		operator = request.user
	usage_events = UsageEvent.objects.filter(operator__is_staff=True, start__gte=first_of_the_month, start__lte=last_of_the_month).exclude(operator=F('user'))
	staff_charges = StaffCharge.objects.filter(start__gte=first_of_the_month, start__lte=last_of_the_month)
	if operator:
		usage_events = usage_events.exclude(~Q(operator_id=operator.id))
		staff_charges = staff_charges.exclude(~Q(staff_member_id=operator.id))
	dictionary = {
		'usage': usage_events,
		'staff_charges': staff_charges,
		'staff_list': User.objects.filter(is_staff=True),
		'month_list': month_list(),
		'selected_staff': operator.id if operator else "all staff",
		'selected_month': request.GET.get('date'),
	}
	return render(request, 'remote_work.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def validate_staff_charge(request, staff_charge_id):
	staff_charge = get_object_or_404(StaffCharge, id=staff_charge_id)
	staff_charge.validated = True
	staff_charge.save()
	return HttpResponse()


@staff_member_required(login_url=None)
@require_POST
def validate_usage_event(request, usage_event_id):
	usage_event = get_object_or_404(UsageEvent, id=usage_event_id)
	usage_event.validated = True
	usage_event.save()
	return HttpResponse()



from django.contrib.admin.views.decorators import staff_member_required
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from NEMO.forms import ConsumableWithdrawForm
from NEMO.models import Consumable, User


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def consumables(request):
	form = ConsumableWithdrawForm(request.POST or None, initial={'quantity': 1})

	dictionary = {
		'users': User.objects.filter(is_active=True),
		'consumables': Consumable.objects.filter(visible=True).order_by('category', 'name'),
	}

	if form.is_valid():
		withdraw = form.save(commit=False)
		withdraw.merchant = request.user
		withdraw.save()
		withdraw.consumable.quantity -= withdraw.quantity
		withdraw.consumable.save()
		dictionary['success'] = 'The withdraw for {} was successfully logged.'.format(withdraw.customer)
		form = ConsumableWithdrawForm(initial={'quantity': 1})
	else:
		if hasattr(form, 'cleaned_data') and 'customer' in form.cleaned_data:
			dictionary['projects'] = form.cleaned_data['customer'].active_projects()

	dictionary['form'] = form
	return render(request, 'consumables.html', dictionary)



from rest_framework.serializers import ModelSerializer

from NEMO.models import User, Project, Account, Reservation, AreaAccessRecord, UsageEvent, Task, TaskHistory, ScheduledOutage, Tool


class UserSerializer(ModelSerializer):
	class Meta:
		model = User
		fields = ('id', 'first_name', 'last_name', 'username', 'date_joined')


class ProjectSerializer(ModelSerializer):
	class Meta:
		model = Project
		fields = ('id', 'name', 'application_identifier', 'active')


class AccountSerializer(ModelSerializer):
	class Meta:
		model = Account
		fields = ('id', 'name', 'active')


class ToolSerializer(ModelSerializer):
	class Meta:
		model = Tool
		fields = '__all__'


class ReservationSerializer(ModelSerializer):
	class Meta:
		model = Reservation
		fields = '__all__'


class UsageEventSerializer(ModelSerializer):
	class Meta:
		model = UsageEvent
		fields = '__all__'


class AreaAccessRecordSerializer(ModelSerializer):
	class Meta:
		model = AreaAccessRecord
		fields = '__all__'


class TaskHistorySerializer(ModelSerializer):
	class Meta:
		model = TaskHistory
		fields = '__all__'


class TaskSerializer(ModelSerializer):
	history = TaskHistorySerializer(many=True, read_only=True)

	class Meta:
		model = Task
		fields = '__all__'


class ScheduledOutageSerializer(ModelSerializer):
	class Meta:
		model = ScheduledOutage
		fields = '__all__'



from datetime import date, timedelta
from time import sleep

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.http import urlencode
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO.models import Area, AreaAccessRecord, Door, PhysicalAccessLog, PhysicalAccessType, Project, User
from NEMO.tasks import postpone
from NEMO.utilities import parse_start_and_end_date
from NEMO.views.customization import get_customization


@staff_member_required(login_url=None)
@require_GET
def area_access(request):
	""" Presents a page that displays audit records for all NanoFab areas. """
	today = timezone.now().strftime('%m/%d/%Y')
	yesterday = (timezone.now() - timedelta(days=1)).strftime('%m/%d/%Y')
	dictionary = {
		'today': reverse('area_access') + '?' + urlencode({'start': today, 'end': today}),
		'yesterday': reverse('area_access') + '?' + urlencode({'start': yesterday, 'end': yesterday}),
	}
	try:
		start, end = parse_start_and_end_date(request.GET['start'], request.GET['end'])
		dictionary['start'] = start
		dictionary['end'] = end
		dictionary['access_records'] = AreaAccessRecord.objects.filter(start__gte=start, start__lt=end, staff_charge=None)
	except:
		pass
	return render(request, 'area_access/area_access.html', dictionary)


@login_required
@permission_required('NEMO.add_areaaccessrecord')
@require_GET
def welcome_screen(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	return render(request, 'area_access/welcome_screen.html', {'area': door.area, 'door': door})


@login_required
@permission_required('NEMO.change_areaaccessrecord')
@require_GET
def farewell_screen(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	return render(request, 'area_access/farewell_screen.html', {'area': door.area, 'door': door})


@login_required
@permission_required('NEMO.add_areaaccessrecord')
@require_POST
def login_to_area(request, door_id):
	door = get_object_or_404(Door, id=door_id)

	badge_number = request.POST.get('badge_number', '')
	if badge_number == '':
		return render(request, 'area_access/badge_not_found.html')
	try:
		badge_number = int(badge_number)
		user = User.objects.get(badge_number=badge_number)
	except (User.DoesNotExist, ValueError):
		return render(request, 'area_access/badge_not_found.html')

	log = PhysicalAccessLog()
	log.user = user
	log.door = door
	log.time = timezone.now()
	log.result = PhysicalAccessType.DENY  # Assume the user does not have access

	# Check if the user is active
	if not user.is_active:
		log.details = "This user is not active, preventing them from entering any access controlled areas."
		log.save()
		return render(request, 'area_access/inactive.html')

	# Check if the user has any physical access levels
	if not user.physical_access_levels.all().exists():
		log.details = "This user does not belong to ANY physical access levels."
		log.save()
		message = "You have not been granted physical access to any NanoFab area. Please visit the User Office if you believe this is an error."
		return render(request, 'area_access/physical_access_denied.html', {'message': message})

	# Check if the user normally has access to this door at the current time
	if not any([access_level.accessible() for access_level in user.physical_access_levels.filter(area=door.area)]):
		log.details = "This user is not assigned to a physical access level that allows access to this door at this time."
		log.save()
		message = "You do not have access to this area of the NanoFab at this time. Please visit the User Office if you believe this is an error."
		return render(request, 'area_access/physical_access_denied.html', {'message': message})

	# Check that the user's physical access has not expired
	if user.access_expiration is not None and user.access_expiration < date.today():
		log.details = "This user was blocked from this physical access level because their physical access has expired."
		log.save()
		message = "Your physical access to the NanoFab has expired. Have you completed your safety training within the last year? Please visit the User Office to renew your access."
		return render(request, 'area_access/physical_access_denied.html', {'message': message})

	# Users may not access an area if a required resource is unavailable.
	# Staff are exempt from this rule.
	unavailable_resources = door.area.required_resources.filter(available=False)
	if unavailable_resources and not user.is_staff:
		log.details = "The user was blocked from entering this area because a required resource was unavailable."
		log.save()
		return render(request, 'area_access/resource_unavailable.html', {'unavailable_resources': unavailable_resources})

	# Users must have at least one billable project in order to enter an area.
	if user.active_project_count() == 0:
		log.details = "The user has no active projects, preventing them from entering an access controlled area."
		log.save()
		return render(request, 'area_access/no_active_projects.html')

	current_area_access_record = user.area_access_record()
	if current_area_access_record and current_area_access_record.area == door.area:
		# No log entry necessary here because all validation checks passed.
		# The log entry is captured when the subsequent choice is made by the user.
		return render(request, 'area_access/already_logged_in.html', {'area': door.area, 'project': current_area_access_record.project, 'badge_number': user.badge_number})

	previous_area = None
	if user.active_project_count() == 1:
		log.result = PhysicalAccessType.ALLOW
		log.save()

		# Automatically log the user out of any previous area before logging them in to the new area.
		if user.in_area():
			previous_area_access_record = user.area_access_record()
			previous_area_access_record.end = timezone.now()
			previous_area_access_record.save()
			previous_area = previous_area_access_record.area

		record = AreaAccessRecord()
		record.area = door.area
		record.customer = user
		record.project = user.active_projects()[0]
		record.save()
		unlock_door(door)
		return render(request, 'area_access/login_success.html', {'area': door.area, 'name': user.first_name, 'project': record.project, 'previous_area': previous_area})
	elif user.active_project_count() > 1:
		project_id = request.POST.get('project_id')
		if project_id:
			project = get_object_or_404(Project, id=project_id)
			if project not in user.active_projects():
				log.details = "The user attempted to bill the project named {}, but they are not a member of that project.".format(project.name)
				log.save()
				message = "You are not authorized to bill this project."
				return render(request, 'area_access/physical_access_denied.html', {'message': message})
			log.result = PhysicalAccessType.ALLOW
			log.save()

			# Automatically log the user out of any previous area before logging them in to the new area.
			if user.in_area():
				previous_area_access_record = user.area_access_record()
				previous_area_access_record.end = timezone.now()
				previous_area_access_record.save()
				previous_area = previous_area_access_record.area

			record = AreaAccessRecord()
			record.area = door.area
			record.customer = user
			record.project = project
			record.save()
			unlock_door(door)
			return render(request, 'area_access/login_success.html', {'area': door.area, 'name': user.first_name, 'project': record.project, 'previous_area': previous_area})
		else:
			# No log entry necessary here because all validation checks passed, and the user must indicate which project
			# the wish to login under. The log entry is captured when the subsequent choice is made by the user.
			return render(request, 'area_access/choose_project.html', {'area': door.area, 'user': user})


@postpone
def unlock_door(door):
	door.interlock.unlock()
	sleep(8)
	door.interlock.lock()
	sleep(3)


@login_required
@permission_required('NEMO.change_areaaccessrecord')
@require_POST
def logout_of_area(request, door_id):
	try:
		badge_number = int(request.POST.get('badge_number', ''))
		user = User.objects.get(badge_number=badge_number)
	except (User.DoesNotExist, ValueError):
		return render(request, 'area_access/badge_not_found.html')
	record = user.area_access_record()
	# Allow the user to log out of any area, even if this is a logout tablet for a different area.
	if record:
		record.end = timezone.now()
		record.save()
		return render(request, 'area_access/logout_success.html', {'area': record.area, 'name': user.first_name})
	else:
		return render(request, 'area_access/not_logged_in.html')


@staff_member_required(login_url=None)
@require_POST
def force_area_logout(request, user_id):
	user = get_object_or_404(User, id=user_id)
	record = user.area_access_record()
	if record is None:
		return HttpResponseBadRequest('That user is not logged into the {}.'.format(record.area))
	record.end = timezone.now()
	record.save()
	return HttpResponse()


@login_required
@permission_required('NEMO.change_areaaccessrecord')
@require_POST
def open_door(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	badge_number = request.POST.get('badge_number', '')
	try:
		badge_number = int(badge_number)
		user = User.objects.get(badge_number=badge_number)
	except (User.DoesNotExist, ValueError):
		return render(request, 'area_access/badge_not_found.html')
	if user.area_access_record() and user.area_access_record().area == door.area:
		log = PhysicalAccessLog(user=user, door=door, time=timezone.now(), result=PhysicalAccessType.ALLOW, details="The user was permitted to enter this area, and already had an active area access record for this area.")
		log.save()
		unlock_door(door)
		return render(request, 'area_access/door_is_open.html')
	return render(request, 'area_access/not_logged_in.html', {'area': door.area})


@login_required
@require_http_methods(['GET', 'POST'])
def change_project(request, new_project=None):
	""" For area access, allow the user to stop billing a project and start billing another project. """
	if request.method == 'GET':
		return render(request, 'area_access/change_project.html')
	old_project = request.user.billing_to_project()
	if old_project is None:
		dictionary = {
			'error': "There was a problem changing the project you're billing for area access. You must already be billing a project in order to change to another project, however you were not logged in to an area."
		}
		return render(request, 'area_access/change_project.html', dictionary)
	# If we're already billing the requested project then there's nothing to do.
	if old_project.id == new_project:
		return redirect(reverse('landing'))
	new_project = get_object_or_404(Project, id=new_project)
	if new_project not in request.user.active_projects():
		dictionary = {
			'error': 'You do not have permission to bill that project.'
		}
		return render(request, 'area_access/change_project.html', dictionary)
	# Stop billing the user's initial project
	record = request.user.area_access_record()
	record.end = timezone.now()
	record.save()
	area = record.area
	# Start billing the user's new project
	record = AreaAccessRecord()
	record.area = area
	record.customer = request.user
	record.project = new_project
	record.save()
	return redirect(reverse('landing'))


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def new_area_access_record(request):
	dictionary = {
		'customers': User.objects.filter(is_active=True)
	}
	if request.method == 'GET':
		try:
			customer = User.objects.get(id=request.GET['customer'], is_active=True)
			dictionary['customer'] = customer
			dictionary['areas'] = list(set([access_level.area for access_level in customer.physical_access_levels.all()]))
			if customer.active_project_count() == 0:
				dictionary['error_message'] = '{} does not have any active projects to bill area access'.format(customer)
				return render(request, 'area_access/new_area_access_record.html', dictionary)
			if not dictionary['areas']:
				dictionary['error_message'] = '{} does not have access to any billable NanoFab areas'.format(customer)
				return render(request, 'area_access/new_area_access_record.html', dictionary)
			return render(request, 'area_access/new_area_access_record_details.html', dictionary)
		except:
			pass
		return render(request, 'area_access/new_area_access_record.html', dictionary)
	if request.method == 'POST':
		try:
			user = User.objects.get(id=request.POST['customer'], is_active=True)
			project = Project.objects.get(id=request.POST['project'])
			area = Area.objects.get(id=request.POST['area'])
		except:
			dictionary['error_message'] = 'Your request contained an invalid identifier.'
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if user.access_expiration is not None and user.access_expiration < timezone.now().date():
			dictionary['error_message'] = '{} does not have access to the {} because the user\'s physical access expired on {}. You must update the user\'s physical access expiration date before creating a new area access record.'.format(user, area.name.lower(), user.access_expiration.strftime('%B %m, %Y'))
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if not any([access_level.accessible() for access_level in user.physical_access_levels.filter(area=area)]):
			dictionary['error_message'] = '{} does not have a physical access level that allows access to the {} at this time.'.format(user, area.name.lower())
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if user.billing_to_project():
			dictionary['error_message'] = '{} is already billing area access to another area. The user must log out of that area before entering another.'.format(user)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if project not in user.active_projects():
			dictionary['error_message'] = '{} is not authorized to bill that project.'.format(user)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if area.required_resources.filter(available=False).exists():
			dictionary['error_message'] = 'The {} is inaccessible because a required resource is unavailable. You must make all required resources for this area available before creating a new area access record.'.format(area.name.lower())
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		record = AreaAccessRecord()
		record.area = area
		record.customer = user
		record.project = project
		record.save()
		dictionary['success'] = '{} is now logged in to the {}.'.format(user, area.name.lower())
		return render(request, 'area_access/new_area_access_record.html', dictionary)


@login_required
@require_http_methods(['GET', 'POST'])
def self_log_in(request):
	if not able_to_self_log_in_to_area(request.user):
		return redirect(reverse('landing'))
	dictionary = {
		'projects': request.user.active_projects(),
	}
	areas = []
	for access_level in request.user.physical_access_levels.all():
		unavailable_resources = access_level.area.required_resources.filter(available=False)
		if access_level.accessible() and not unavailable_resources:
			areas.append(access_level.area)
	dictionary['areas'] = areas
	if request.method == 'GET':
		return render(request, 'area_access/self_login.html', dictionary)
	if request.method == 'POST':
		try:
			a = Area.objects.get(id=request.POST['area'])
			p = Project.objects.get(id=request.POST['project'])
			if a in dictionary['areas'] and p in dictionary['projects']:
				AreaAccessRecord.objects.create(area=a, customer=request.user, project=p)
		except:
			pass
		return redirect(reverse('landing'))


def able_to_self_log_in_to_area(user):
	# 'Self log in' must be enabled
	if not get_customization('self_log_in') == 'enabled':
		return False
	# Check if the user is active
	if not user.is_active:
		return False
	# Check if the user has a billable project
	if user.active_project_count() < 1:
		return False
	# Check if the user is already in an area. If so, the /change_project/ URL can be used to change their project.
	if user.in_area():
		return False
	# Check if the user has any physical access levels
	if not user.physical_access_levels.all().exists():
		return False
	# Check if the user normally has access to this area at the current time
	accessible_areas = []
	for access_level in user.physical_access_levels.all():
		if access_level.accessible():
			accessible_areas.append(access_level.area)
	if not accessible_areas:
		return False
	# Check that the user's physical access has not expired
	if user.access_expiration is not None and user.access_expiration < date.today():
		return False
	# Staff are exempt from the remaining rule checks
	if user.is_staff:
		return True
	# Users may not access an area if a required resource is unavailable,
	# so return true if there exists at least one area they are able to log in to.
	for area in accessible_areas:
		unavailable_resources = area.required_resources.filter(available=False)
		if not unavailable_resources:
			return True
	# No areas are accessible...
	return False



from atexit import register
from logging import getLogger
from queue import Queue
from threading import Thread

postponed_tasks_logger = getLogger("NEMO.PostponedTasks")


def postpone(function):
	def decorator(*arguments, **named_arguments):
		_queue.put((function, arguments, named_arguments))
	return decorator


def _worker():
	while True:
		function, arguments, named_arguments = _queue.get()
		try:
			function(*arguments, **named_arguments)
		except:
			from traceback import format_exc
			details = format_exc()
			postponed_tasks_logger.exception(details)
		finally:
			_queue.task_done()  # So we can join at exit


def _cleanup():
	_queue.join()  # So we don't exit too soon


_queue = Queue()
_thread = Thread(target=_worker, name='Postponed execution thread')
_thread.daemon = True
_thread.start()
register(_cleanup)



from django import forms
from django.contrib import admin
from django.contrib.admin import register
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import Permission

from NEMO.actions import lock_selected_interlocks, synchronize_with_tool_usage, unlock_selected_interlocks
from NEMO.models import Account, ActivityHistory, Alert, Area, AreaAccessRecord, Comment, Configuration, ConfigurationHistory, Consumable, ConsumableCategory, ConsumableWithdraw, ContactInformation, ContactInformationCategory, Customization, Door, Interlock, InterlockCard, LandingPageChoice, MembershipHistory, News, Notification, PhysicalAccessLevel, PhysicalAccessLog, Project, Reservation, Resource, ResourceCategory, SafetyIssue, ScheduledOutage, ScheduledOutageCategory, StaffCharge, Task, TaskCategory, TaskHistory, TaskStatus, Tool, TrainingSession, UsageEvent, User, UserType

admin.site.site_header = "NEMO"
admin.site.site_title = "NEMO"
admin.site.index_title = "Detailed administration"


def record_remote_many_to_many_changes_and_save(request, obj, form, change, many_to_many_field, save_function_pointer):
	"""
	Record the changes in a many-to-many field that the model does not own. Then, save the many-to-many field.
	"""
	# If the model object is being changed then we can get the list of previous members.
	if change:
		original_members = set(obj.user_set.all())
	else:  # The model object is being created (instead of changed) so we can assume there are no members (initially).
		original_members = set()
	current_members = set(form.cleaned_data[many_to_many_field])
	added_members = []
	removed_members = []

	# Log membership changes if they occurred.
	symmetric_difference = original_members ^ current_members
	if symmetric_difference:
		if change:  # the members have changed, so find out what was added and removed...
			# We can can see the previous members of the object model by looking it up
			# in the database because the member list hasn't been committed yet.
			added_members = set(current_members) - set(original_members)
			removed_members = set(original_members) - set(current_members)

		else:  # a model object is being created (instead of changed) so we can assume all the members are new...
			added_members = form.cleaned_data[many_to_many_field]

	# A primary key for the object is required to make many-to-many field changes.
	# If the object is being changed then it has already been assigned a primary key.
	if not change:
		save_function_pointer(request, obj, form, change)
	obj.user_set = form.cleaned_data[many_to_many_field]
	save_function_pointer(request, obj, form, change)

	# Record which members were added to the object.
	for user in added_members:
		new_member = MembershipHistory()
		new_member.authorizer = request.user
		new_member.parent_content_object = obj
		new_member.child_content_object = user
		new_member.action = MembershipHistory.Action.ADDED
		new_member.save()

	# Record which members were removed from the object.
	for user in removed_members:
		ex_member = MembershipHistory()
		ex_member.authorizer = request.user
		ex_member.parent_content_object = obj
		ex_member.child_content_object = user
		ex_member.action = MembershipHistory.Action.REMOVED
		ex_member.save()


def record_local_many_to_many_changes(request, obj, form, many_to_many_field):
	"""
	Record the changes in a many-to-many field that the model owns.
	"""
	if many_to_many_field in form.changed_data:
		original_members = set(getattr(obj, many_to_many_field).all())
		current_members = set(form.cleaned_data[many_to_many_field])
		added_members = set(current_members) - set(original_members)
		for a in added_members:
			p = MembershipHistory()
			p.action = MembershipHistory.Action.ADDED
			p.authorizer = request.user
			p.child_content_object = obj
			p.parent_content_object = a
			p.save()
		removed_members = set(original_members) - set(current_members)
		for a in removed_members:
			p = MembershipHistory()
			p.action = MembershipHistory.Action.REMOVED
			p.authorizer = request.user
			p.child_content_object = obj
			p.parent_content_object = a
			p.save()


def record_active_state(request, obj, form, field_name, is_initial_creation):
	"""
	Record whether the account, project, or user is active when the active state is changed.
	"""
	if field_name in form.changed_data or is_initial_creation:
		activity_entry = ActivityHistory()
		activity_entry.authorizer = request.user
		activity_entry.action = getattr(obj, field_name)
		activity_entry.content_object = obj
		activity_entry.save()


class ToolAdminForm(forms.ModelForm):
	class Meta:
		model = Tool
		fields = '__all__'

	qualified_users = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	backup_owners = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	required_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Required resources',
			is_stacked=False
		)
	)

	nonrequired_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Nonrequired resources',
			is_stacked=False
		)
	)

	def __init__(self, *args, **kwargs):
		super(ToolAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields['qualified_users'].initial = self.instance.user_set.all()
			self.fields['required_resources'].initial = self.instance.required_resource_set.all()
			self.fields['nonrequired_resources'].initial = self.instance.nonrequired_resource_set.all()


@register(Tool)
class ToolAdmin(admin.ModelAdmin):
	list_display = ('name', 'category', 'visible', 'operational', 'problematic', 'is_configurable')
	list_filter = ('visible', 'operational', 'category')
	form = ToolAdminForm
	fieldsets = (
		(None, {'fields': ('name', 'category', 'qualified_users', 'post_usage_questions'),}),
		('Current state', {'fields': ('visible', 'operational'),}),
		('Contact information', {'fields': ('primary_owner', 'backup_owners', 'notification_email_address', 'location', 'phone_number'),}),
		('Usage policy', {'fields': ('reservation_horizon', 'minimum_usage_block_time', 'maximum_usage_block_time', 'maximum_reservations_per_day', 'minimum_time_between_reservations', 'maximum_future_reservation_time', 'missed_reservation_threshold', 'requires_area_access', 'grant_physical_access_level_upon_qualification', 'grant_badge_reader_access_upon_qualification', 'interlock', 'allow_delayed_logoff'),}),
		('Dependencies', {'fields': ('required_resources', 'nonrequired_resources'),}),
	)

	def save_model(self, request, obj, form, change):
		"""
		Explicitly record any project membership changes.
		"""
		record_remote_many_to_many_changes_and_save(request, obj, form, change, 'qualified_users', super(ToolAdmin, self).save_model)
		if 'required_resources' in form.changed_data:
			obj.required_resource_set = form.cleaned_data['required_resources']
		if 'nonrequired_resources' in form.changed_data:
			obj.nonrequired_resource_set = form.cleaned_data['nonrequired_resources']


@register(TrainingSession)
class TrainingSessionAdmin(admin.ModelAdmin):
	list_display = ('id', 'trainer', 'trainee', 'tool', 'project', 'type', 'date', 'duration', 'qualified')
	list_filter = ('qualified', 'date', 'type', 'tool')
	date_hierarchy = 'date'


@register(StaffCharge)
class StaffChargeAdmin(admin.ModelAdmin):
	list_display = ('id', 'staff_member', 'customer', 'start', 'end')
	list_filter = ('start',)
	date_hierarchy = 'start'


@register(AreaAccessRecord)
class AreaAccessRecordAdmin(admin.ModelAdmin):
	list_display = ('id', 'customer', 'area', 'project', 'start', 'end')
	list_filter = ('area', 'start',)
	date_hierarchy = 'start'


@register(Configuration)
class ConfigurationAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'name', 'qualified_users_are_maintainers', 'display_priority', 'exclude_from_configuration_agenda')
	filter_horizontal = ('maintainers',)


@register(ConfigurationHistory)
class ConfigurationHistoryAdmin(admin.ModelAdmin):
	list_display = ('id', 'configuration', 'user', 'modification_time', 'slot')
	date_hierarchy = 'modification_time'


@register(Account)
class AccountAdmin(admin.ModelAdmin):
	list_display = ('name', 'id', 'active')
	search_fields = ('name',)
	list_filter = ('active',)

	def save_model(self, request, obj, form, change):
		""" Audit account and project active status. """
		super(AccountAdmin, self).save_model(request, obj, form, change)
		record_active_state(request, obj, form, 'active', not change)


class ProjectAdminForm(forms.ModelForm):
	class Meta:
		model = Project
		fields = '__all__'

	members = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	def __init__(self, *args, **kwargs):
		super(ProjectAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields['members'].initial = self.instance.user_set.all()


@register(Project)
class ProjectAdmin(admin.ModelAdmin):
	list_display = ('name', 'id', 'application_identifier', 'account', 'active')
	search_fields = ('name', 'application_identifier', 'account__name')
	list_filter = ('active',)
	form = ProjectAdminForm

	def save_model(self, request, obj, form, change):
		"""
		Audit project creation and modification. Also save any project membership changes explicitly.
		"""
		record_remote_many_to_many_changes_and_save(request, obj, form, change, 'members', super(ProjectAdmin, self).save_model)
		# Make a history entry if a project has been moved under an account.
		# This applies to newly created projects and project ownership reassignment.
		if 'account' in form.changed_data:
			# Create a membership removal entry for the project if it used to belong to another account:
			if change:
				previous_account = MembershipHistory()
				previous_account.authorizer = request.user
				previous_account.child_content_object = obj
				previous_account.parent_content_object = Account.objects.get(pk=form.initial['account'])
				previous_account.action = MembershipHistory.Action.REMOVED
				previous_account.save()

			# Create a membership addition entry for the project with its current account.
			current_account = MembershipHistory()
			current_account.authorizer = request.user
			current_account.child_content_object = obj
			current_account.parent_content_object = obj.account
			current_account.action = MembershipHistory.Action.ADDED
			current_account.save()

		# Record whether the project is active or not.
		record_active_state(request, obj, form, 'active', not change)


@register(Reservation)
class ReservationAdmin(admin.ModelAdmin):
	list_display = ('id', 'user', 'creator', 'tool', 'project', 'start', 'end', 'duration', 'cancelled', 'missed')
	list_filter = ('cancelled', 'missed', 'tool')
	date_hierarchy = 'start'


@register(UsageEvent)
class UsageEventAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'user', 'operator', 'project', 'start', 'end', 'duration')
	list_filter = ('start', 'end', 'tool')
	date_hierarchy = 'start'


@register(Consumable)
class ConsumableAdmin(admin.ModelAdmin):
	list_display = ('name', 'quantity', 'category', 'visible', 'reminder_threshold', 'reminder_email')
	list_filter = ('visible', 'category')


@register(ConsumableCategory)
class ConsumableCategoryAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(ConsumableWithdraw)
class ConsumableWithdrawAdmin(admin.ModelAdmin):
	list_display = ('id', 'customer', 'merchant', 'consumable', 'quantity', 'project', 'date')
	list_filter = ('date', 'consumable')
	date_hierarchy = 'date'


@register(InterlockCard)
class InterlockCardAdmin(admin.ModelAdmin):
	list_display = ('server', 'port', 'number', 'even_port', 'odd_port')


@register(Interlock)
class InterlockAdmin(admin.ModelAdmin):
	list_display = ('id', 'card', 'channel', 'state', 'tool', 'door')
	actions = [lock_selected_interlocks, unlock_selected_interlocks, synchronize_with_tool_usage]
	readonly_fields = ['state', 'most_recent_reply']


@register(Task)
class TaskAdmin(admin.ModelAdmin):
	list_display = ('id', 'urgency', 'tool', 'creator', 'creation_time', 'problem_category', 'cancelled', 'resolved', 'resolution_category')
	list_filter = ('urgency', 'resolved', 'cancelled', 'safety_hazard', 'creation_time', 'tool')
	date_hierarchy = 'creation_time'


@register(TaskCategory)
class TaskCategoryAdmin(admin.ModelAdmin):
	list_display = ('name', 'stage')


@register(TaskStatus)
class TaskStatusAdmin(admin.ModelAdmin):
	list_display = ('name', 'notify_primary_tool_owner', 'notify_backup_tool_owners', 'notify_tool_notification_email', 'custom_notification_email_address')


@register(TaskHistory)
class TaskHistoryAdmin(admin.ModelAdmin):
	list_display = ('id', 'task', 'status', 'time', 'user')
	readonly_fields = ('time',)
	date_hierarchy = 'time'


@register(Comment)
class CommentAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'author', 'creation_date', 'expiration_date', 'visible', 'hidden_by', 'hide_date')
	list_filter = ('visible', 'creation_date', 'tool')
	date_hierarchy = 'creation_date'
	search_fields = ('content',)


@register(Resource)
class ResourceAdmin(admin.ModelAdmin):
	list_display = ('name', 'category', 'available')
	list_filter = ('available', 'category')
	filter_horizontal = ('fully_dependent_tools', 'partially_dependent_tools', 'dependent_areas')


@register(ActivityHistory)
class ActivityHistoryAdmin(admin.ModelAdmin):
	list_display = ('__str__', 'content_type', 'object_id', 'action', 'date', 'authorizer')
	date_hierarchy = 'date'


@register(MembershipHistory)
class MembershipHistoryAdmin(admin.ModelAdmin):
	list_display = ('__str__', 'parent_content_type', 'parent_object_id', 'action', 'child_content_type', 'child_object_id', 'date', 'authorizer')
	date_hierarchy = 'date'


@register(UserType)
class UserTypeAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(User)
class UserAdmin(admin.ModelAdmin):
	filter_horizontal = ('groups', 'user_permissions', 'qualifications', 'projects', 'physical_access_levels')
	fieldsets = (
		('Personal information', {'fields': ('first_name', 'last_name', 'username', 'email', 'badge_number', 'type', 'domain')}),
		('Permissions', {'fields': ('is_active', 'is_staff', 'is_technician', 'is_superuser', 'training_required', 'groups', 'user_permissions', 'physical_access_levels')}),
		('Important dates', {'fields': ('date_joined', 'last_login', 'access_expiration')}),
		('NanoFab information', {'fields': ('qualifications', 'projects')}),
	)
	search_fields = ('first_name', 'last_name', 'username', 'email')
	list_display = ('first_name', 'last_name', 'username', 'email', 'is_active', 'domain', 'is_staff', 'is_technician', 'is_superuser', 'date_joined', 'last_login')
	list_filter = ('is_active', 'domain', 'is_staff', 'is_technician', 'is_superuser', 'date_joined', 'last_login')

	def save_model(self, request, obj, form, change):
		""" Audit project membership and qualifications when a user is saved. """
		super(UserAdmin, self).save_model(request, obj, form, change)
		record_local_many_to_many_changes(request, obj, form, 'projects')
		record_local_many_to_many_changes(request, obj, form, 'qualifications')
		record_local_many_to_many_changes(request, obj, form, 'physical_access_levels')
		record_active_state(request, obj, form, 'is_active', not change)


@register(PhysicalAccessLog)
class PhysicalAccessLogAdmin(admin.ModelAdmin):
	list_display = ('user', 'door', 'time', 'result')
	list_filter = ('door', 'result')
	search_fields = ('user',)
	date_hierarchy = 'time'


@register(SafetyIssue)
class SafetyIssueAdmin(admin.ModelAdmin):
	list_display = ('id', 'reporter', 'creation_time', 'visible', 'resolved', 'resolution_time', 'resolver')
	list_filter = ('resolved', 'visible', 'creation_time', 'resolution_time')
	readonly_fields = ('creation_time', 'resolution_time')
	search_fields = ('location', 'concern', 'progress', 'resolution',)


@register(Door)
class DoorAdmin(admin.ModelAdmin):
	list_display = ('name', 'area', 'interlock', 'get_absolute_url')


@register(Alert)
class AlertAdmin(admin.ModelAdmin):
	list_display = ('title', 'creation_time', 'creator', 'debut_time', 'expiration_time', 'user', 'dismissible')


@register(PhysicalAccessLevel)
class PhysicalAccessLevelAdmin(admin.ModelAdmin):
	list_display = ('name', 'area', 'schedule')


@register(ContactInformationCategory)
class ContactInformationCategoryAdmin(admin.ModelAdmin):
	list_display = ('name', 'display_order')


@register(ContactInformation)
class ContactInformationAdmin(admin.ModelAdmin):
	list_display = ('name', 'category')


@register(LandingPageChoice)
class LandingPageChoiceAdmin(admin.ModelAdmin):
	list_display = ('display_priority', 'name', 'url', 'open_in_new_tab', 'secure_referral', 'hide_from_mobile_devices', 'hide_from_desktop_computers')
	list_display_links = ('name',)


@register(Customization)
class CustomizationAdmin(admin.ModelAdmin):
	list_display = ('name', 'value')


@register(ScheduledOutageCategory)
class ScheduledOutageCategoryAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(ScheduledOutage)
class ScheduledOutageAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'resource', 'creator', 'title', 'start', 'end')


@register(News)
class NewsAdmin(admin.ModelAdmin):
	list_display = ('id', 'created', 'last_updated', 'archived', 'title')
	list_filter = ('archived',)


@register(Notification)
class NotificationAdmin(admin.ModelAdmin):
	list_display = ('id', 'user', 'expiration', 'content_type', 'object_id')


admin.site.register(ResourceCategory)
admin.site.register(Area)
admin.site.register(Permission)



from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.core.mail import send_mail
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.template import Context, Template
from django.urls import reverse
from django.views.decorators.http import require_GET, require_http_methods

from NEMO.forms import SafetyIssueCreationForm, SafetyIssueUpdateForm
from NEMO.models import SafetyIssue
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.notifications import create_safety_notification, delete_safety_notification, get_notifications


@login_required
@require_http_methods(['GET', 'POST'])
def safety(request):
	dictionary = {}
	if request.method == 'POST':
		form = SafetyIssueCreationForm(request.user, data=request.POST)
		if form.is_valid():
			issue = form.save()
			send_safety_email_notification(request, issue)
			dictionary['title'] = 'Concern received'
			dictionary['heading'] = 'Your safety concern was sent to NanoFab staff and will be addressed promptly'
			create_safety_notification(issue)
			return render(request, 'acknowledgement.html', dictionary)
	tickets = SafetyIssue.objects.filter(resolved=False).order_by('-creation_time')
	if not request.user.is_staff:
		tickets = tickets.filter(visible=True)
	dictionary['tickets'] = tickets
	dictionary['safety_introduction'] = get_media_file_contents('safety_introduction.html')
	dictionary['notifications'] = get_notifications(request.user, SafetyIssue)
	return render(request, 'safety/safety.html', dictionary)


def send_safety_email_notification(request, issue):
	subject = 'Safety issue'
	dictionary = {
		'issue': issue,
		'issue_absolute_url': request.build_absolute_uri(issue.get_absolute_url()),
	}
	recipient = get_customization('safety_email_address')
	message = get_media_file_contents('safety_issue_email.html')
	if not recipient or not message:
		return
	rendered_message = Template(message).render(Context(dictionary))
	from_email = issue.reporter.email if issue.reporter else recipient
	send_mail(subject, '', from_email, [recipient], html_message=rendered_message)


@login_required
@require_GET
def resolved_safety_issues(request):
	tickets = SafetyIssue.objects.filter(resolved=True)
	if not request.user.is_staff:
		tickets = tickets.filter(visible=True)
	return render(request, 'safety/resolved_issues.html', {'tickets': tickets})


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def update_safety_issue(request, ticket_id):
	if request.method == 'POST':
		ticket = get_object_or_404(SafetyIssue, id=ticket_id)
		form = SafetyIssueUpdateForm(request.user, data=request.POST, instance=ticket)
		if form.is_valid():
			issue = form.save()
			if issue.resolved:
				delete_safety_notification(issue)
			return HttpResponseRedirect(reverse('safety'))
	dictionary = {
		'ticket': get_object_or_404(SafetyIssue, id=ticket_id)
	}
	return render(request, 'safety/update_issue.html', dictionary)



from django.contrib.admin.views.decorators import staff_member_required
from django.core.files.storage import get_storage_class
from django.core.validators import validate_email
from django.http import HttpResponseBadRequest
from django.shortcuts import redirect, render
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import Customization


def get_media_file_contents(file_name):
	""" Get the contents of a media file if it exists. Return a blank string if it does not exist. """
	storage = get_storage_class()()
	if not storage.exists(file_name):
		return ''
	f = storage.open(file_name)
	return f.read().strip()


def store_media_file(content, file_name):
	""" Delete any existing media file with the same name and save the new content into file_name in the media directory. If content is blank then no new file is created. """
	storage = get_storage_class()()
	storage.delete(file_name)
	if content:
		storage.save(file_name, content)


customizable_key_values = [
	'feedback_email_address',
	'user_office_email_address',
	'safety_email_address',
	'abuse_email_address',
	'self_log_in',
]

customizable_content = [
	'login_banner',
	'safety_introduction',
	'nanofab_rules_tutorial',
	'cancellation_email',
	'feedback_email',
	'generic_email',
	'missed_reservation_email',
	'nanofab_rules_tutorial_email',
	'new_task_email',
	'reservation_reminder_email',
	'reservation_warning_email',
	'safety_issue_email',
	'staff_charge_reminder_email',
	'task_status_notification',
	'unauthorized_tool_access_email',
	'usage_reminder_email',
]


def get_customization(name):
	if name not in customizable_key_values:
		raise Exception('Invalid customization')
	try:
		return Customization.objects.get(name=name).value
	except Customization.DoesNotExist:
		return ''


def set_customization(name, value):
	if name not in customizable_key_values:
		raise Exception(f'Invalid customization: {value}')
	if value:
		if name in ['feedback_email_address', 'user_office_email_address', 'safety_email_address', 'abuse_email_address']:
			validate_email(value)
		Customization.objects.update_or_create(name=name, defaults={'value': value})
	else:
		try:
			Customization.objects.get(name=name).delete()
		except Customization.DoesNotExist:
			pass


@staff_member_required(login_url=None)
@require_GET
def customization(request):
	dictionary = {x: get_media_file_contents(x + '.html') for x in customizable_content}
	dictionary.update({y: get_customization(y) for y in customizable_key_values})
	return render(request, 'customizations.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def customize(request, element):
	if element in customizable_content:
		store_media_file(request.FILES.get(element, ''), element + '.html')
	elif element == 'email_addresses':
		set_customization('feedback_email_address', request.POST.get('feedback_email_address', ''))
		set_customization('safety_email_address', request.POST.get('safety_email_address', ''))
		set_customization('abuse_email_address', request.POST.get('abuse_email_address', ''))
		set_customization('user_office_email_address', request.POST.get('user_office_email_address', ''))
	elif element == 'application_settings':
		set_customization('self_log_in', request.POST.get('self_log_in', ''))
	else:
		return HttpResponseBadRequest('Invalid customization')
	return redirect('customization')



def lock_selected_interlocks(model_admin, request, queryset):
	for interlock in queryset:
		interlock.lock()


def unlock_selected_interlocks(model_admin, request, queryset):
	for interlock in queryset:
		interlock.unlock()


def synchronize_with_tool_usage(model_admin, request, queryset):
	for interlock in queryset:
		# Ignore interlocks with no tool assigned, and ignore interlocks connected to doors
		if not interlock.tool or interlock.door:
			continue
		if interlock.tool.in_use():
			interlock.unlock()
		else:
			interlock.lock()



from datetime import timedelta
from http import HTTPStatus
from urllib.parse import urljoin

import requests
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest
from django.shortcuts import render, redirect, get_object_or_404
from django.utils import timezone
from django.views.decorators.http import require_GET, require_http_methods, require_POST, logger

from NEMO.admin import record_local_many_to_many_changes, record_active_state
from NEMO.forms import UserForm
from NEMO.models import User, Project, Tool, PhysicalAccessLevel, Reservation, StaffCharge, UsageEvent, AreaAccessRecord, ActivityHistory


@staff_member_required(login_url=None)
@require_GET
def users(request):
	all_users = User.objects.all()
	return render(request, 'users/users.html', {'users': all_users})


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def create_or_modify_user(request, user_id):
	dictionary = {
		'projects': Project.objects.filter(active=True, account__active=True),
		'tools': Tool.objects.filter(visible=True),
		'physical_access_levels': PhysicalAccessLevel.objects.all(),
		'one_year_from_now': timezone.now() + timedelta(days=365),
		'identity_service_available': settings.IDENTITY_SERVICE['available'],
		'identity_service_domains': settings.IDENTITY_SERVICE['domains'],
	}
	try:
		user = User.objects.get(id=user_id)
	except:
		user = None

	if dictionary['identity_service_available']:
		try:
			result = requests.get(urljoin(settings.IDENTITY_SERVICE['url'], '/areas/'), timeout=3)
			if result.status_code == HTTPStatus.OK:
				dictionary['externally_managed_physical_access_levels'] = result.json()
			else:
				dictionary['identity_service_available'] = False
				warning_message = 'The identity service encountered a problem while attempting to return a list of externally managed areas. The NEMO administrator has been notified to resolve the problem.'
				dictionary['warning'] = warning_message
				warning_message += ' The HTTP error was {}: {}'.format(result.status_code, result.text)
				logger.error(warning_message)
		except Exception as e:
			dictionary['identity_service_available'] = False
			warning_message = 'There was a problem communicating with the identity service. NEMO is unable to retrieve the list of externally managed areas. The NEMO administrator has been notified to resolve the problem.'
			dictionary['warning'] = warning_message
			warning_message += ' An exception was encountered: ' + type(e).__name__ + ' - ' + str(e)
			logger.error(warning_message)
	else:
		dictionary['warning'] = 'The identity service is disabled. You will not be able to modify externally managed physical access levels, reset account passwords, or unlock accounts.'

	if request.method == 'GET':
		dictionary['form'] = UserForm(instance=user)
		try:
			if dictionary['identity_service_available'] and user and user.is_active and user.domain:
				parameters = {
					'username': user.username,
					'domain': user.domain,
				}
				result = requests.get(settings.IDENTITY_SERVICE['url'], parameters, timeout=3)
				if result.status_code == HTTPStatus.OK:
					dictionary['user_identity_information'] = result.json()
				elif result.status_code == HTTPStatus.NOT_FOUND:
					dictionary['warning'] = "The identity service could not find username {} on the {} domain. Does the user's account reside on a different domain? If so, select that domain now and save the user information.".format(user.username, user.domain)
				else:
					dictionary['identity_service_available'] = False
					warning_message = 'The identity service encountered a problem while attempting to search for a user. The NEMO administrator has been notified to resolve the problem.'
					dictionary['warning'] = warning_message
					warning_message += ' The HTTP error was {}: {}'.format(result.status_code, result.text)
					logger.error(warning_message)
		except Exception as e:
			dictionary['identity_service_available'] = False
			warning_message = 'There was a problem communicating with the identity service. NEMO is unable to search for a user. The NEMO administrator has been notified to resolve the problem.'
			dictionary['warning'] = warning_message
			warning_message += ' An exception was encountered: ' + type(e).__name__ + ' - ' + str(e)
			logger.error(warning_message)
		return render(request, 'users/create_or_modify_user.html', dictionary)
	elif request.method == 'POST':
		form = UserForm(request.POST, instance=user)
		dictionary['form'] = form
		if not form.is_valid():
			return render(request, 'users/create_or_modify_user.html', dictionary)

		# Remove the user account from the domain if it's deactivated, changed domain, or changed username...
		if dictionary['identity_service_available'] and user:
			no_longer_active = form.initial['is_active'] is True and form.cleaned_data['is_active'] is False
			domain_switched = form.initial['domain'] != '' and form.initial['domain'] != form.cleaned_data['domain']
			username_changed = form.initial['username'] != form.cleaned_data['username']
			if no_longer_active or domain_switched or username_changed:
				parameters = {
					'username': form.initial['username'],
					'domain': form.initial['domain'],
				}
				try:
					result = requests.delete(settings.IDENTITY_SERVICE['url'], data=parameters, timeout=3)
					# If the delete succeeds, or the user is not found, then everything is ok.
					if result.status_code not in (HTTPStatus.OK, HTTPStatus.NOT_FOUND):
						dictionary['identity_service_available'] = False
						logger.error('The identity service encountered a problem while attempting to delete a user. The HTTP error is {}: {}'.format(result.status_code, result.text))
						dictionary['warning'] = 'The user information was not modified because the identity service could not delete the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
						return render(request, 'users/create_or_modify_user.html', dictionary)
				except Exception as e:
					dictionary['identity_service_available'] = False
					logger.error('There was a problem communicating with the identity service while attempting to delete a user. An exception was encountered: ' + type(e).__name__ + ' - ' + str(e))
					dictionary['warning'] = 'The user information was not modified because the identity service could not delete the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
					return render(request, 'users/create_or_modify_user.html', dictionary)

		# Ensure the user account is added and configured correctly on the current domain if the user is active...
		if dictionary['identity_service_available'] and form.cleaned_data['is_active']:
			parameters = {
				'username': form.cleaned_data['username'],
				'domain': form.cleaned_data['domain'],
				'badge_number': form.cleaned_data.get('badge_number', ''),
				'email': form.cleaned_data.get('email'),
				'access_expiration': form.cleaned_data.get('access_expiration'),
				'requested_areas': request.POST.getlist('externally_managed_access_levels'),
			}
			try:
				if len(parameters['requested_areas']) > 0 and not parameters['badge_number']:
					dictionary['warning'] = 'A user must have a badge number in order to have area access. Please enter the badge number first, then grant access to areas.'
					return render(request, 'users/create_or_modify_user.html', dictionary)
				result = requests.put(settings.IDENTITY_SERVICE['url'], data=parameters, timeout=3)
				if result.status_code == HTTPStatus.NOT_FOUND:
					dictionary['warning'] = 'The username was not found on this domain. Did you spell the username correctly in this form and did you select the correct domain? Ensure the user exists on the domain in order to proceed.'
					return render(request, 'users/create_or_modify_user.html', dictionary)
				if result.status_code != HTTPStatus.OK:
					dictionary['identity_service_available'] = False
					logger.error('The identity service encountered a problem while attempting to modify a user. The HTTP error is {}: {}'.format(result.status_code, result.text))
					dictionary['warning'] = 'The user information was not modified because the identity service encountered a problem while creating the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
					return render(request, 'users/create_or_modify_user.html', dictionary)
			except Exception as e:
				dictionary['identity_service_available'] = False
				logger.error('There was a problem communicating with the identity service while attempting to modify a user. An exception was encountered: ' + type(e).__name__ + ' - ' + str(e))
				dictionary['warning'] = 'The user information was not modified because the identity service encountered a problem while creating the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
				return render(request, 'users/create_or_modify_user.html', dictionary)

		# Only save the user model for now, and wait to process the many-to-many relationships.
		# This way, many-to-many changes can be recorded.
		# See this web page for more information:
		# https://docs.djangoproject.com/en/dev/topics/forms/modelforms/#the-save-method
		user = form.save(commit=False)
		user.save()
		record_active_state(request, user, form, 'is_active', user_id == 'new')
		record_local_many_to_many_changes(request, user, form, 'qualifications')
		record_local_many_to_many_changes(request, user, form, 'physical_access_levels')
		record_local_many_to_many_changes(request, user, form, 'projects')
		form.save_m2m()

		return redirect('users')
	else:
		return HttpResponseBadRequest('Invalid method')


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def deactivate(request, user_id):
	dictionary = {
		'user_to_deactivate': get_object_or_404(User, id=user_id),
		'reservations': Reservation.objects.filter(user=user_id, cancelled=False, missed=False, end__gt=timezone.now()),
		'staff_charges': StaffCharge.objects.filter(customer=user_id, end=None),
		'tool_usage': UsageEvent.objects.filter(user=user_id, end=None).prefetch_related('tool'),
	}
	user_to_deactivate = dictionary['user_to_deactivate']
	if request.method == 'GET':
		return render(request, 'users/safe_deactivation.html', dictionary)
	elif request.method == 'POST':
		if settings.IDENTITY_SERVICE['available']:
			parameters = {
				'username': user_to_deactivate.username,
				'domain': user_to_deactivate.domain,
			}
			try:
				result = requests.delete(settings.IDENTITY_SERVICE['url'], data=parameters, timeout=3)
				# If the delete succeeds, or the user is not found, then everything is ok.
				if result.status_code not in (HTTPStatus.OK, HTTPStatus.NOT_FOUND):
					logger.error(f'The identity service encountered a problem while attempting to delete a user. The HTTP error is {result.status_code}: {result.text}')
					dictionary['warning'] = 'The user information was not modified because the identity service could not delete the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
					return render(request, 'users/safe_deactivation.html', dictionary)
			except Exception as e:
				logger.error('There was a problem communicating with the identity service while attempting to delete a user. An exception was encountered: ' + type(e).__name__ + ' - ' + str(e))
				dictionary['warning'] = 'The user information was not modified because the identity service could not delete the corresponding domain account. The NEMO administrator has been notified to resolve the problem.'
				return render(request, 'users/safe_deactivation.html', dictionary)

		if request.POST.get('cancel_reservations') == 'on':
			# Cancel all reservations that haven't ended
			for reservation in dictionary['reservations']:
				reservation.cancelled = True
				reservation.cancellation_time = timezone.now()
				reservation.cancelled_by = request.user
				reservation.save()
		if request.POST.get('disable_tools') == 'on':
			# End all current tool usage
			for usage_event in dictionary['tool_usage']:
				if usage_event.tool.interlock and not usage_event.tool.interlock.lock():
					error_message = f"The interlock command for the {usage_event.tool} failed. The error message returned: {usage_event.tool.interlock.most_recent_reply}"
					logger.error(error_message)
				usage_event.end = timezone.now()
				usage_event.save()
		if request.POST.get('force_area_logout') == 'on':
			area_access = user_to_deactivate.area_access_record()
			if area_access:
				area_access.end = timezone.now()
				area_access.save()
		if request.POST.get('end_staff_charges') == 'on':
			# End a staff charge that the user might be performing
			staff_charge = user_to_deactivate.get_staff_charge()
			if staff_charge:
				staff_charge.end = timezone.now()
				staff_charge.save()
				try:
					area_access = AreaAccessRecord.objects.get(staff_charge=staff_charge, end=None)
					area_access.end = timezone.now()
					area_access.save()
				except AreaAccessRecord.DoesNotExist:
					pass
			# End all staff charges that are being performed for the user
			for staff_charge in dictionary['staff_charges']:
				staff_charge.end = timezone.now()
				staff_charge.save()
				try:
					area_access = AreaAccessRecord.objects.get(staff_charge=staff_charge, end=None)
					area_access.end = timezone.now()
					area_access.save()
				except AreaAccessRecord.DoesNotExist:
					pass
		user_to_deactivate.is_active = False
		user_to_deactivate.save()
		activity_entry = ActivityHistory()
		activity_entry.authorizer = request.user
		activity_entry.action = ActivityHistory.Action.DEACTIVATED
		activity_entry.content_object = user_to_deactivate
		activity_entry.save()
		return redirect('users')


@staff_member_required(login_url=None)
@require_POST
def reset_password(request, user_id):
	try:
		user = get_object_or_404(User, id=user_id)
		result = requests.post(urljoin(settings.IDENTITY_SERVICE['url'], '/reset_password/'), {'username': user.username, 'domain': user.domain}, timeout=3)
		if result.status_code == HTTPStatus.OK:
			dictionary = {
				'title': 'Password reset',
				'heading': 'The account password was set to the default',
			}
		else:
			dictionary = {
				'title': 'Oops',
				'heading': 'There was a problem resetting the password',
				'content': 'The identity service returned HTTP error code {}. {}'.format(result.status_code, result.text),
			}
	except Exception as e:
		dictionary = {
			'title': 'Oops',
			'heading': 'There was a problem communicating with the identity service',
			'content': 'Exception caught: {}. {}'.format(type(e).__name__, str(e)),
		}
	return render(request, 'acknowledgement.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def unlock_account(request, user_id):
	try:
		user = get_object_or_404(User, id=user_id)
		result = requests.post(urljoin(settings.IDENTITY_SERVICE['url'], '/unlock_account/'), {'username': user.username, 'domain': user.domain}, timeout=3)
		if result.status_code == HTTPStatus.OK:
			dictionary = {
				'title': 'Account unlocked',
				'heading': 'The account is now unlocked',
			}
		else:
			dictionary = {
				'title': 'Oops',
				'heading': 'There was a problem unlocking the account',
				'content': 'The identity service returned HTTP error code {}. {}'.format(result.status_code, result.text),
			}
	except Exception as e:
		dictionary = {
			'title': 'Oops',
			'heading': 'There was a problem communicating with the identity service',
			'content': 'Exception caught: {}. {}'.format(type(e).__name__, str(e)),
		}
	return render(request, 'acknowledgement.html', dictionary)



from datetime import timedelta

from django import template
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse, NoReverseMatch
from django.utils import timezone
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def class_name(value):
	return value.__class__.__name__


@register.filter
def is_soon(time):
	""" 'Soon' is defined as within 10 minutes from now. """
	return time <= timezone.now() + timedelta(minutes=10)


@register.filter()
def to_int(value):
	return int(value)


@register.filter
def json_search_base(items_to_search):
	result = '['
	for item in items_to_search:
		result += '{{"name":"{0}", "id":{1}}},'.format(escape(str(item)), item.id)
	result = result.rstrip(',') + ']'
	return mark_safe(result)


@register.simple_tag
def json_search_base_with_extra_fields(items_to_search, *extra_fields):
	"""
	This tag is similar to the json_search_base filter, but adds extra information upon request.
	The type of object is always provided in the JSON output. Thus, you have a heterogeneous collection
	of objects and differentiate them in your JavaScript. The extra fields are only added when an
	object actually has that attribute. Otherwise, the code skips over the request.
	"""
	result = '['
	for item in items_to_search:
		object_type = ContentType.objects.get_for_model(item).name
		result += '{{"name":"{0}", "id":{1}, "type":"{2}"'.format(escape(str(item)), item.id, object_type)
		for x in extra_fields:
			if hasattr(item, x):
				result += ', "{0}":"{1}"'.format(x, getattr(item, x))
		result += '},'
	result = result.rstrip(',') + ']'
	return mark_safe(result)


@register.simple_tag
def navigation_url(url_name, description):
	try:
		return format_html('<li><a href="{}">{}</a></li>', reverse(url_name), description)
	except NoReverseMatch:
		return ''


@register.filter
def get_item(dictionary, key):
	return dictionary.get(key)



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_GET
from requests import get

from NEMO.models import AreaAccessRecord, ConsumableWithdraw, Reservation, StaffCharge, TrainingSession, UsageEvent, User
from NEMO.utilities import get_month_timeframe, month_list


@login_required
@require_GET
def usage(request):
	first_of_the_month, last_of_the_month = get_month_timeframe(request.GET.get('timeframe'))
	dictionary = {
		'area_access': AreaAccessRecord.objects.filter(customer=request.user, end__gt=first_of_the_month, end__lte=last_of_the_month),
		'consumables': ConsumableWithdraw.objects.filter(customer=request.user, date__gt=first_of_the_month, date__lte=last_of_the_month),
		'missed_reservations': Reservation.objects.filter(user=request.user, missed=True, end__gt=first_of_the_month, end__lte=last_of_the_month),
		'staff_charges': StaffCharge.objects.filter(customer=request.user, end__gt=first_of_the_month, end__lte=last_of_the_month),
		'training_sessions': TrainingSession.objects.filter(trainee=request.user, date__gt=first_of_the_month, date__lte=last_of_the_month),
		'usage_events': UsageEvent.objects.filter(user=request.user, end__gt=first_of_the_month, end__lte=last_of_the_month),
		'month_list': month_list(),
		'timeframe': request.GET.get('timeframe') or first_of_the_month.strftime('%B, %Y'),
	}
	dictionary['no_charges'] = not (dictionary['area_access'] or dictionary['consumables'] or dictionary['missed_reservations'] or dictionary['staff_charges'] or dictionary['training_sessions'] or dictionary['usage_events'])
	return render(request, 'usage/usage.html', dictionary)


@login_required
@require_GET
def billing_information(request, timeframe=''):
	dictionary = {}
	if not settings.BILLING_SERVICE['available']:
		return HttpResponse()
	try:
		cost_activity_url = settings.BILLING_SERVICE['cost_activity_url']
		project_lead_url = settings.BILLING_SERVICE['project_lead_url']
		keyword_arguments = settings.BILLING_SERVICE['keyword_arguments']

		first_of_the_month, last_of_the_month = get_month_timeframe(timeframe)
		formatted_projects = ','.join(map(str, set(request.user.active_projects().values_list('application_identifier', flat=True))))
		cost_activity_params = {
			'created_date_gte': f"'{first_of_the_month.strftime('%m/%d/%Y')}'",
			'created_date_lt': f"'{last_of_the_month.strftime('%m/%d/%Y')}'",
			'application_names': f"'{formatted_projects}'",
			'$format': 'json'
		}
		cost_activity_response = get(cost_activity_url, params=cost_activity_params, **keyword_arguments)
		cost_activity_data = cost_activity_response.json()['d']

		latest_pis_params = {'$format': 'json'}
		latest_pis_response = get(project_lead_url, params=latest_pis_params, **keyword_arguments)
		latest_pis_data = latest_pis_response.json()['d']

		project_totals = {}
		application_totals = {}
		user_pi_applications = list()
		# Construct a tree of account, application, project, and member total spending
		cost_activities_tree = {}
		for activity in cost_activity_data:
			project_totals.setdefault(activity['project_id'], 0)
			application_totals.setdefault(activity['application_id'], 0)
			account_key = (activity['account_id'], activity['account_name'])
			application_key = (activity['application_id'], activity['application_name'])
			project_key = (activity['project_id'], activity['project_name'])
			user_key = (activity['member_id'], User.objects.get(id=activity['member_id']))
			user_is_pi = is_user_pi(request.user, next((x for x in latest_pis_data if x['application_name'] == activity['application_name']), None))
			if user_is_pi:
				user_pi_applications.append(activity['application_id'])
			if user_is_pi or str(request.user.id) == activity['member_id']:
				cost_activities_tree.setdefault((activity['account_id'], activity['account_name']), {})
				cost_activities_tree[account_key].setdefault(application_key, {})
				cost_activities_tree[account_key][application_key].setdefault(project_key, {})
				cost_activities_tree[account_key][application_key][project_key].setdefault(user_key, 0)
				cost = -activity['cost'] if activity['activity_type'] == 'refund_activity' else activity['cost']
				cost_activities_tree[account_key][application_key][project_key][user_key] = cost_activities_tree[account_key][application_key][project_key][user_key] + cost
				project_totals[activity['project_id']] = project_totals[activity['project_id']] + cost
				application_totals[activity['application_id']] = application_totals[activity['application_id']] + cost
		dictionary['spending'] = {
			'activities': cost_activities_tree,
			'project_totals': project_totals,
			'application_totals': application_totals,
			'user_pi_applications': user_pi_applications
		} if cost_activities_tree else {}
		return render(request, 'usage/billing.html', dictionary)
	except Exception as e:
		return HttpResponse(str(e))


def is_user_pi(user, application_pi_row):
	return application_pi_row is not None and (user.username == application_pi_row['username'] or (user.first_name == application_pi_row['first_name'] and user.last_name == application_pi_row['last_name']))



from django.contrib.auth.decorators import login_required
from django.core.mail import send_mail
from django.shortcuts import render
from django.template import Template, RequestContext, Context
from django.views.decorators.http import require_http_methods

from NEMO.models import User, Project
from NEMO.views.customization import get_customization, get_media_file_contents


@login_required
@require_http_methods(['GET', 'POST'])
def nanofab_rules(request):
	if request.method == 'GET':
		tutorial = get_media_file_contents('nanofab_rules_tutorial.html')
		if tutorial:
			dictionary = {
				'active_user_count': User.objects.filter(is_active=True).count(),
				'active_project_count': Project.objects.filter(active=True).count(),
			}
			tutorial = Template(tutorial).render(RequestContext(request, dictionary))
		return render(request, 'nanofab_rules.html', {'nanofab_rules_tutorial': tutorial})
	elif request.method == 'POST':
		summary = request.POST.get('making_reservations_summary', '').strip()[:3000]
		dictionary = {
			'user': request.user,
			'making_reservations_rule_summary': summary,
		}
		abuse_email = get_customization('abuse_email_address')
		email_contents = get_media_file_contents('nanofab_rules_tutorial_email.html')
		if abuse_email and email_contents:
			message = Template(email_contents, dictionary).render(Context(dictionary))
			send_mail('NanoFab rules tutorial', '', abuse_email, [abuse_email], html_message=message)
		dictionary = {
			'title': 'NanoFab rules tutorial',
			'heading': 'Tutorial complete!',
			'content': 'Tool usage and reservation privileges have been enabled on your user account.',
		}
		request.user.training_required = False
		request.user.save()
		return render(request, 'acknowledgement.html', dictionary)



from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import User, StaffCharge, AreaAccessRecord, Project, Area


@staff_member_required(login_url=None)
@require_GET
def staff_charges(request):
	staff_charge = request.user.get_staff_charge()
	if staff_charge:
		try:
			area_access_record = AreaAccessRecord.objects.get(staff_charge=staff_charge.id, end=None)
			return render(request, 'staff_charges/end_area_charge.html', {'area': area_access_record.area})
		except AreaAccessRecord.DoesNotExist:
			return render(request, 'staff_charges/change_status.html', {'areas': Area.objects.all()})
	error = None
	customer = None
	try:
		customer = User.objects.get(id=request.GET['customer'])
	except:
		pass
	if customer:
		if customer.active_project_count() > 0:
			return render(request, 'staff_charges/choose_project.html', {'customer': customer})
		else:
			error = str(customer) + ' does not have any active projects. You cannot bill staff time to this user.'
	users = User.objects.filter(is_active=True).exclude(id=request.user.id)
	return render(request, 'staff_charges/new_staff_charge.html', {'users': users, 'error': error})


@staff_member_required(login_url=None)
@require_POST
def begin_staff_charge(request):
	if request.user.charging_staff_time():
		return HttpResponseBadRequest('You cannot create a new staff charge when one is already in progress.')
	charge = StaffCharge()
	charge.customer = User.objects.get(id=request.POST['customer'])
	charge.project = Project.objects.get(id=request.POST['project'])
	charge.staff_member = request.user
	charge.save()
	return redirect(reverse('staff_charges'))


@staff_member_required(login_url=None)
@require_POST
def end_staff_charge(request):
	if not request.user.charging_staff_time():
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot end it.')
	charge = request.user.get_staff_charge()
	charge.end = timezone.now()
	charge.save()
	try:
		area_access = AreaAccessRecord.objects.get(staff_charge=charge, end=None)
		area_access.end = timezone.now()
		area_access.save()
	except AreaAccessRecord.DoesNotExist:
		pass
	return redirect(reverse('staff_charges'))


@staff_member_required(login_url=None)
@require_POST
def begin_staff_area_charge(request):
	charge = request.user.get_staff_charge()
	if not charge:
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot begin an area access charge.')
	if AreaAccessRecord.objects.filter(staff_charge=charge, end=None).count() > 0:
		return HttpResponseBadRequest('You cannot create an area access charge when one is already in progress.')
	try:
		area = Area.objects.get(id=request.POST['area'])
	except:
		return HttpResponseBadRequest('Invalid area')
	area_access = AreaAccessRecord()
	area_access.area = area
	area_access.staff_charge = charge
	area_access.customer = charge.customer
	area_access.project = charge.project
	area_access.save()
	return redirect(reverse('staff_charges'))


@staff_member_required(login_url=None)
@require_POST
def end_staff_area_charge(request):
	charge = request.user.get_staff_charge()
	if not charge:
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot end area access.')
	area_access = AreaAccessRecord.objects.get(staff_charge=charge, end=None)
	area_access.end = timezone.now()
	area_access.save()
	return redirect(reverse('staff_charges'))



from copy import deepcopy
from datetime import timedelta
from http import HTTPStatus
from itertools import chain

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseNotFound, HttpResponseServerError
from django.shortcuts import get_object_or_404, redirect, render
from django.utils import timezone
from django.views.decorators.http import logger, require_GET, require_POST

from NEMO.forms import CommentForm, nice_errors
from NEMO.models import Comment, Configuration, ConfigurationHistory, Project, Reservation, StaffCharge, Task, TaskCategory, TaskStatus, Tool, UsageEvent, User
from NEMO.utilities import extract_times, quiet_int
from NEMO.views.policy import check_policy_to_disable_tool, check_policy_to_enable_tool
from NEMO.widgets.dynamic_form import DynamicForm
from NEMO.widgets.tool_tree import ToolTree


@login_required
@require_GET
def tool_control(request, tool_id=None):
	""" Presents the tool control view to the user, allowing them to being/end using a tool or see who else is using it. """
	if request.user.active_project_count() == 0:
		return render(request, 'no_project.html')
	# The tool-choice sidebar is not available for mobile devices, so redirect the user to choose a tool to view.
	if request.device == 'mobile' and tool_id is None:
		return redirect('choose_tool', next_page='tool_control')
	tools = Tool.objects.filter(visible=True).order_by('category', 'name')
	dictionary = {
		'tools': tools,
		'selected_tool': tool_id,
	}
	# The tool-choice sidebar only needs to be rendered for desktop devices, not mobile devices.
	if request.device == 'desktop':
		dictionary['rendered_tool_tree_html'] = ToolTree().render(None, {'tools': tools})
	return render(request, 'tool_control/tool_control.html', dictionary)


@login_required
@require_GET
def tool_status(request, tool_id):
	""" Gets the current status of the tool (that is, whether it is currently in use or not). """
	tool = get_object_or_404(Tool, id=tool_id, visible=True)

	dictionary = {
		'tool': tool,
		'task_categories': TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT),
		'rendered_configuration_html': tool.configuration_widget(request.user),
		'mobile': request.device == 'mobile',
		'task_statuses': TaskStatus.objects.all(),
		'post_usage_questions': DynamicForm(tool.post_usage_questions).render(),
	}

	try:
		current_reservation = Reservation.objects.get(start__lt=timezone.now(), end__gt=timezone.now(), cancelled=False, missed=False, shortened=False, user=request.user, tool=tool)
		if request.user == current_reservation.user:
			dictionary['time_left'] = current_reservation.end
	except Reservation.DoesNotExist:
		pass

	# Staff need the user list to be able to qualify users for the tool.
	if request.user.is_staff:
		dictionary['users'] = User.objects.filter(is_active=True)

	return render(request, 'tool_control/tool_status.html', dictionary)


@staff_member_required(login_url=None)
@require_GET
def use_tool_for_other(request):
	dictionary = {
		'users': User.objects.filter(is_active=True).exclude(id=request.user.id)
	}
	return render(request, 'tool_control/use_tool_for_other.html', dictionary)


@login_required
@require_POST
def tool_configuration(request):
	""" Sets the current configuration of a tool. """
	try:
		configuration = Configuration.objects.get(id=request.POST['configuration_id'])
	except:
		return HttpResponseNotFound('Configuration not found.')
	if configuration.tool.in_use():
		return HttpResponseBadRequest('Cannot change a configuration while a tool is in use.')
	if not configuration.user_is_maintainer(request.user):
		return HttpResponseBadRequest('You are not authorized to change this configuration.')
	try:
		slot = int(request.POST['slot'])
		choice = int(request.POST['choice'])
	except:
		return HttpResponseBadRequest('Invalid configuration parameters.')
	try:
		configuration.replace_current_setting(slot, choice)
	except IndexError:
		return HttpResponseBadRequest('Invalid configuration choice.')
	configuration.save()
	history = ConfigurationHistory()
	history.configuration = configuration
	history.slot = slot
	history.user = request.user
	history.setting = configuration.get_current_setting(slot)
	history.save()
	return HttpResponse()


@login_required
@require_POST
def create_comment(request):
	form = CommentForm(request.POST)
	if not form.is_valid():
		return HttpResponseBadRequest(nice_errors(form).as_ul())
	comment = form.save(commit=False)
	comment.content = comment.content.strip()
	comment.author = request.user
	comment.expiration_date = None if form.cleaned_data['expiration'] == 0 else timezone.now() + timedelta(days=form.cleaned_data['expiration'])
	comment.save()
	return redirect('tool_control')


@login_required
@require_POST
def hide_comment(request, comment_id):
	comment = get_object_or_404(Comment, id=comment_id)
	if comment.author_id != request.user.id and not request.user.is_staff:
		return HttpResponseBadRequest("You may only hide a comment if you are its author or a staff member.")
	comment.visible = False
	comment.hidden_by = request.user
	comment.hide_date = timezone.now()
	comment.save()
	return redirect('tool_control')


def determine_tool_status(tool):
	# Make the tool operational when all problems are resolved that require a shutdown.
	if tool.task_set.filter(force_shutdown=True, cancelled=False, resolved=False).count() == 0:
		tool.operational = True
	else:
		tool.operational = False
	tool.save()


@login_required
@require_POST
def enable_tool(request, tool_id, user_id, project_id, staff_charge):
	""" Enable a tool for a user. The user must be qualified to do so based on the lab usage policy. """

	if not settings.ALLOW_CONDITIONAL_URLS:
		return HttpResponseBadRequest('Tool control is only available on campus. We\'re working to change that! Thanks for your patience.')

	tool = get_object_or_404(Tool, id=tool_id)
	operator = request.user
	user = get_object_or_404(User, id=user_id)
	project = get_object_or_404(Project, id=project_id)
	staff_charge = staff_charge == 'true'
	response = check_policy_to_enable_tool(tool, operator, user, project, staff_charge)
	if response.status_code != HTTPStatus.OK:
		return response

	# All policy checks passed so enable the tool for the user.
	if tool.interlock and not tool.interlock.unlock():
		raise Exception("The interlock command for this tool failed. The error message returned: " + str(tool.interlock.most_recent_reply))

	# Create a new usage event to track how long the user uses the tool.
	new_usage_event = UsageEvent()
	new_usage_event.operator = operator
	new_usage_event.user = user
	new_usage_event.project = project
	new_usage_event.tool = tool
	new_usage_event.save()

	if staff_charge:
		new_staff_charge = StaffCharge()
		new_staff_charge.staff_member = request.user
		new_staff_charge.customer = user
		new_staff_charge.project = project
		new_staff_charge.save()

	return response


@login_required
@require_POST
def disable_tool(request, tool_id):

	if not settings.ALLOW_CONDITIONAL_URLS:
		return HttpResponseBadRequest('Tool control is only available on campus.')

	tool = get_object_or_404(Tool, id=tool_id)
	if tool.get_current_usage_event() is None:
		return HttpResponse()
	downtime = timedelta(minutes=quiet_int(request.POST.get('downtime')))
	response = check_policy_to_disable_tool(tool, request.user, downtime)
	if response.status_code != HTTPStatus.OK:
		return response
	try:
		current_reservation = Reservation.objects.get(start__lt=timezone.now(), end__gt=timezone.now(), cancelled=False, missed=False, shortened=False, user=request.user, tool=tool)
		# Staff are exempt from mandatory reservation shortening when tool usage is complete.
		if request.user.is_staff is False:
			# Shorten the user's reservation to the current time because they're done using the tool.
			new_reservation = deepcopy(current_reservation)
			new_reservation.id = None
			new_reservation.pk = None
			new_reservation.end = timezone.now() + downtime
			new_reservation.save()
			current_reservation.shortened = True
			current_reservation.descendant = new_reservation
			current_reservation.save()
	except Reservation.DoesNotExist:
		pass

	# All policy checks passed so disable the tool for the user.
	if tool.interlock and not tool.interlock.lock():
		error_message = f"The interlock command for the {tool} failed. The error message returned: {tool.interlock.most_recent_reply}"
		logger.error(error_message)
		return HttpResponseServerError(error_message)

	# End the current usage event for the tool
	current_usage_event = tool.get_current_usage_event()
	current_usage_event.end = timezone.now() + downtime

	# Collect post-usage questions
	dynamic_form = DynamicForm(tool.post_usage_questions)
	current_usage_event.run_data = dynamic_form.extract(request)
	dynamic_form.charge_for_consumable(current_usage_event.user, current_usage_event.operator, current_usage_event.project, current_usage_event.run_data)

	current_usage_event.save()
	if request.user.charging_staff_time():
		existing_staff_charge = request.user.get_staff_charge()
		if existing_staff_charge.customer == current_usage_event.user and existing_staff_charge.project == current_usage_event.project:
			response = render(request, 'staff_charges/reminder.html', {'tool': tool})

	return response


@login_required
@require_GET
def past_comments_and_tasks(request):
	try:
		start, end = extract_times(request.GET)
	except:
		return HttpResponseBadRequest('Please enter a start and end date.')
	tool_id = request.GET.get('tool_id')
	try:
		tasks = Task.objects.filter(tool_id=tool_id, creation_time__gt=start, creation_time__lt=end)
		comments = Comment.objects.filter(tool_id=tool_id, creation_date__gt=start, creation_date__lt=end)
	except:
		return HttpResponseBadRequest('Task and comment lookup failed.')
	past = list(chain(tasks, comments))
	past.sort(key=lambda x: getattr(x, 'creation_time', None) or getattr(x, 'creation_date', None))
	past.reverse()
	dictionary = {
		'past': past,
	}
	return render(request, 'tool_control/past_tasks_and_comments.html', dictionary)


@login_required
@require_GET
def ten_most_recent_past_comments_and_tasks(request, tool_id):
	tasks = Task.objects.filter(tool_id=tool_id).order_by('-creation_time')[:10]
	comments = Comment.objects.filter(tool_id=tool_id).order_by('-creation_date')[:10]
	past = list(chain(tasks, comments))
	past.sort(key=lambda x: getattr(x, 'creation_time', None) or getattr(x, 'creation_date', None))
	past.reverse()
	past = past[0:10]
	dictionary = {
		'past': past,
	}
	return render(request, 'tool_control/past_tasks_and_comments.html', dictionary)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-08-15 15:18
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0001_version_1_0_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='ScheduledOutage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField()),
                ('end', models.DateTimeField()),
                ('title', models.CharField(max_length=100)),
                ('details', models.TextField(blank=True)),
                ('creator', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('tool', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool')),
            ],
        ),
        migrations.AddField(
            model_name='landingpagechoice',
            name='hide_from_users',
            field=models.BooleanField(default=False, help_text='Hides this choice from normal users. When checked, only staff, technicians, and super-users can see the choice'),
        ),
    ]
